#                           PUBLIC DOMAIN NOTICE
#              National Center for Biotechnology Information
#
# This software is a "United States Government Work" under the
# terms of the United States Copyright Act.  It was written as part of
# the authors' official duties as United States Government employees and
# thus cannot be copyrighted.  This software is freely available
# to the public for use.  The National Library of Medicine and the U.S.
# Government have not placed any restriction on its use or reproduction.
#
# Although all reasonable efforts have been taken to ensure the accuracy
# and reliability of the software and data, the NLM and the U.S.
# Government do not and cannot warrant the performance or results that
# may be obtained by using this software or data.  The NLM and the U.S.
# Government disclaim all warranties, express or implied, including
# warranties of performance, merchantability or fitness for any particular
# purpose.
#
# Please cite NCBI in any work or product based on this material.

"""
elb/base.py - Definitions of types used by different elb modules

Author: Greg Boratyn (boratyng@ncbi.nlm.nih.gov)
Created: Tue 09 Feb 2021 03:52:31 PM EDT
"""

import configparser
import re
from dataclasses import dataclass, field, fields, _MISSING_TYPE
from enum import Enum, auto
from typing import Dict, List, Union, Optional, NamedTuple

@dataclass(frozen=True)
class InstanceProperties:
    """Properties of a cloud instance

    Attributes:
        ncpus: Number of vCPUs
        memory: Available RAM
    """
    ncpus : int
    memory : float


@dataclass(frozen=True)
class QuerySplittingResults:
    """ Results from query splitting operation """
    query_length : int
    query_batches : List[str]

    def num_batches(self) -> int:
        return len(self.query_batches)


class PositiveInteger(int):
    """A subclass of int that only accepts positive integers. The value is
    validated before object creation"""
    # a bug in mypy does not allow for type annotation here:
    # https://github.com/python/mypy/issues/6061
    def __new__(cls, value):
        """Constructor, validates that argumant is a positive integer after
        conversion to int"""
        try:
            if isinstance(value, float) and value != round(value):
                raise ValueError
            int_value = int(value)
            if int_value <= 0:
                raise ValueError()
        except ValueError as err:
            msg = 'Must be a positive integer.'
            raise ValueError(msg)
        return super(cls, cls).__new__(cls, value)


class Percentage(int):
    """A subclass of int that accepts only percentages: an integer between 0
    and 100"""
    def __new__(cls, value):
        try:
            if isinstance(value, float) and value != round(value):
                raise ValueError
            int_value = int(value)
            if int_value < 0 or int_value > 100:
                raise ValueError
        except ValueError:
            msg = 'Percentage must be a positive integer between 0 and 100'
            raise ValueError(msg)
        return super(cls, cls).__new__(cls, value)


class BoolFromStr:
    """A class that converts strings to boolean values.
    False is created if string value is one of: n, no, 0, false, or empty
    string. True is created otherwise. String values are not case sensitive."""
    def __new__(cls, value):
        if isinstance(value, str):
            return not value.lower() in ['n', 'no', '0', 'false', '']
        return bool(value)


class MemoryStr(str):
    """A subclass of str that only accepts properly formated memory amounts as
    a number followed by a single chanracter for a unit. The value is
    validated before object creation"""
    def __new__(cls, value):
        """Constructor, validates that argumant is a valid GCP name"""
        str_value = str(value)
        number_re = re.compile(r'^\d+[kKmMgGtT]$|^\d+.\d+[kKmMgGtT]$')
        if not number_re.match(str_value):
            raise ValueError('Memory request or limit must be specified by a number followed by a unit, for example 100m')
        if float(str_value[:-1]) <= 0:
            raise ValueError('Memory request or limit must be larger than zero')
        return super(cls, cls).__new__(cls, str_value)


    def asGB(self) -> float:
        """Return the amount of memory in GB as float"""
        mult = 1.0
        if self[-1].upper() == 'K':
            mult /= 1024 ** 2
        elif self[-1].upper() == 'M':
            mult /= 1024
        elif self[-1].upper() == 'T':
            mult *= 1024
        return float(self[:-1]) * mult

    def asMB(self) -> float:
        """Return the amount of memory in MB as float"""
        mult = 1.0
        if self[-1].upper() == 'K':
            mult /= 1024
        elif self[-1].upper() == 'G':
            mult *= 1024
        elif self[-1].upper() == 'T':
            mult *= 1024 ** 2
        return float(self[:-1]) * mult


class DBSource(Enum):
    """Sources of a BLAST database supported by update_blastdb.pl from BLAST+ package"""
    GCP = auto()
    AWS = auto()
    NCBI = auto()
    def __repr__(self):
        return f"'{self.name}'"


class ParamInfo(NamedTuple):
    """Data structure used to link config parameters with ConfigParser
    parameter names:
        section: ConfigParser section name
        param_name: ConfigParser parameter name"""
    #FIXME: add environment and command line parameter names
    section: str
    param_name: str


@dataclass
class ConfigParserToDataclassMapper:
    """Base class that provides methods for dataclasses with elements linked
    to ConfigParser parameter names. A child class must be a dataclass.

    Attributes:
        mapping: A dictionary with a map (class attribute, ConfigParser
        parameter or None)
    """
    
    mapping: Dict[str, Optional[ParamInfo]] = field(init=False)

    def __init__(self):
        """Contructor needed so that dataclass does not auto generate one"""
        pass


    @classmethod
    def create_from_cfg(cls, parser, **kwargs):
        """Meant to be used by a subclass. Create a subclass object
        initializing attribute values from ConfigParser object using mapping
        in the mapped dictonary.

        Arguments:
            parser: A ConfigParser object
            kwargs: Other parameters required by sublcass constructor

        Raises:
            ValueError: if a required parameter is missing in ConfigParser or
            ValueError is raised during subclass attribute initialization
        """
        # check that all dataclass attributes are mapped to configparser params
        cls.validate_mapping()
        errors = []
        for field in fields(cls):
            if field.init:
                mapped = cls.mapping[field.name]
                # ignore dataclass attributes mapped to None
                if mapped is None:
                    continue
                # skip dataclass attributes that with default values and no
                # parameter values in ConfigParser object
                # field.default == dataclass._MISSING_TYPE means that the
                # dataclass attribute has no default value
                if not isinstance(field.default, _MISSING_TYPE) and \
                       (mapped.section not in parser or \
                        mapped.param_name not in parser[mapped.section]):
                    continue

                # report a required parameter missing in ConfigParser
                if isinstance(field.default, _MISSING_TYPE) and \
                       (mapped.section not in parser or \
                        mapped.param_name not in parser[mapped.section]):
                    errors.append(f'Missing {mapped.param_name}')
                    continue

                # initialize dataclass attribute value, call the appropriate
                # class constructor
                kwargs[field.name] = cls.initialize_value(field, mapped,
                                                          parser, errors)

        # report attribute initialization errors
        if errors:
            # if errors were reported for required parameters, test optional
            # parameters and add include error messages before raising
            # exception
            cls._test_optional_cfg_params(parser, errors)
            raise ValueError('\n'.join(errors))

        # call subclass constructor
        obj = cls(**kwargs)
        # initialize parameters not settable via class constructor
        obj._init_optional_from_cfg(parser, errors)
        if errors:
            raise ValueError('\n'.join(errors))
        return obj

    
    @classmethod
    def validate_mapping(cls):
        """Verify that all class attributes appear in the mapping dictionary.
        Raises AttributeError if an attribute is not in mapping."""
        for field in fields(cls):
            if field.name not in cls.mapping and field.name != 'mapping':
                raise AttributeError(f'Field {field.name} does not have mapping to ConfigParser params')


    def _init_optional_from_cfg(self, parser: configparser.ConfigParser,
                                errors: List[str]):
        """Initialize from ConfigParser only class attributes that are not
        parameters of class constructor.

        Parameters:
            parser: ConfigParser object
            errors: A list where error messages will be appended
        """
        self.validate_mapping()
        for field in fields(self):
            # skip attrubutes initialized via class constructor and those that
            # map to None
            if field.name == 'mapping' or field.init or self.mapping[field.name] is None:
                continue
            mapped = self.mapping[field.name]
            if mapped is None:
                continue
            if mapped.section in parser and mapped.param_name in parser[mapped.section]:
                param = self.initialize_value(field, mapped, parser, errors)
                self.__setattr__(field.name, param)


    @classmethod
    def _test_optional_cfg_params(cls, parser: configparser.ConfigParser,
                                  errors: List[str]):
        """Try initializing from ConfigParser only class attributes that are
        not parameters of class constructor to report any inavlid values

        Parameters:
            parser: ConfigParser object
            errors: A list where error messages will be appended
        """
        cls.validate_mapping()
        for field in fields(cls):
            # skip attrubutes initialized via class constructor and those that
            # map to None
            if field.name == 'mapping' or field.init or cls.mapping[field.name] is None:
                continue
            mapped = cls.mapping[field.name]
            if mapped is None:
                continue
            if mapped.section in parser and mapped.param_name in parser[mapped.section]:
                cls.initialize_value(field, mapped, parser, errors)


    @classmethod
    def initialize_value(cls, field, mapped, parser, errors):
        """Helper function to initialize a single attribute from a
        ConfigParser object parameter value.

        Attributes:
            field: dataclass field object
            mapped: ParamInfo object
            parser: ConfigParser object
            errors: Error messages will be added to thie list
        """
        # dataclass attribute type
        ftype = field.type
        # Union types cannot be initialized, types like Optional[int] are
        # really Union[int, None]. If the attribute type is a Union,
        # initialize the first type that is not None.
        # FIXME: in python3.8 this can be done via typing.get_origin()
        if getattr(ftype, '__origin__', None) is not None and \
               ftype.__origin__ == Union:
            ftype = [t for t in ftype.__args__ if t != type(None)][0]

        # if attribute type is an enum, initialize it from str
        if issubclass(ftype, Enum):
            try:
                # first try the string as is
                value = ftype[parser[mapped.section][mapped.param_name]]
            except KeyError:
                # then try uppercase
                try:
                    value = ftype[parser[mapped.section][mapped.param_name].upper()]
                except KeyError:
                    errors.append(f'Parameter "{mapped.param_name}" has invalid value: "{parser[mapped.section][mapped.param_name]}", should be one of {", ".join([i.name for i in ftype])}')
                    # in case of an error initialize the attribute to any value
                    # so that something can be returned and problems with
                    # more parameters can be reported from a single run
                    value = [i for i in ftype][0]

        # for boolean attributes, use ConfigParser converter from str
        elif ftype == bool:
            value = parser.getboolean(mapped.section, mapped.param_name)

        # otherwise call attribute's class constructor with ConfigParser
        # parameter value
        else:
            try:
                value = ftype(parser[mapped.section][mapped.param_name])
            except ValueError as err:
                errors.append(f'Parameter "{mapped.param_name}" has an invalid value: "{parser[mapped.section][mapped.param_name]}": {str(err)}')
                value = None
        return value


    # FIXME: this function does not really belong in this class and should
    # be part of another class or a class decorator
    def __setattr__(self, name, value):
        """Prevent creation of new attributes to catch misspelled class
        attribute values. Raises AttributeError if a value is being assigned to
        a new class attribute."""
        if name not in [i.name for i in fields(self)]:
            raise AttributeError(f'Attribute {name} does not exit in class {type(self)}')
        super().__setattr__(name, value)


    def __getattr__(self, name):
        """Return None for uninitialized dataclass attributes.
        Raises AttrubuteError for other non-existant class attributes"""
        if name in [i.name for i in fields(self)]:
            return None
        else:
            raise AttributeError(f'"{type(self).__name__}" has no attribute "{name}"')



#                           PUBLIC DOMAIN NOTICE
#              National Center for Biotechnology Information
#  
# This software is a "United States Government Work" under the
# terms of the United States Copyright Act.  It was written as part of
# the authors' official duties as United States Government employees and
# thus cannot be copyrighted.  This software is freely available
# to the public for use.  The National Library of Medicine and the U.S.
# Government have not placed any restriction on its use or reproduction.
#   
# Although all reasonable efforts have been taken to ensure the accuracy
# and reliability of the software and data, the NLM and the U.S.
# Government do not and cannot warrant the performance or results that
# may be obtained by using this software or data.  The NLM and the U.S.
# Government disclaim all warranties, express or implied, including
# warranties of performance, merchantability or fitness for any particular
# purpose.
#   
# Please cite NCBI in any work or product based on this material.

"""
elb/commands/status.py - check status of job execution

Author: Victor Joukov joukovv@ncbi.nlm.nih.gov
"""

import sys
import time
import logging
from elastic_blast import gcp
from elastic_blast import aws
from elastic_blast.util import SafeExecError, UserReportError
from elastic_blast.status import get_status
from elastic_blast.constants import CLUSTER_ERROR, CSP, ElbCommand
from elastic_blast.elb_config import ElasticBlastConfig

def create_arg_parser(subparser, common_opts_parser):
    """ Create the command line options subparser for the status command. """
    parser = subparser.add_parser('status', parents=[common_opts_parser],
                                  help='Get the status of an ElasticBLAST search')
    parser.add_argument("--wait", action='store_true',
                        help="Wait for job completion")
    parser.add_argument('--verbose', default=False, action='store_true',
                        help='Detailed information about jobs')
    # FIXME: EB-132
    parser.add_argument("--run-label", type=str,
                        help="Run-label for this ElasticBLAST search, format: key:value")
    parser.set_defaults(func=_status)


#TODO: use cfg only when args.wait, args.sync, and args.run_label are replicated in cfg
def _status(args, cfg, clean_up_stack):
    """ Entry point to handle checking status for an ElasticBLAST search """
    cfg.validate(ElbCommand.STATUS)
    returncode = 0
    try:
        dry_run = cfg.cluster.dry_run

        if cfg.cloud_provider.cloud == CSP.GCP:
            cfg.appstate.k8s_ctx = gcp.get_gke_credentials(cfg)

        verbose_result = ''
        while True:

            if cfg.cloud_provider.cloud == CSP.AWS:
                eb = aws.ElasticBlastAws(cfg)
                counts, verbose_result = eb.check_status(args.verbose)
                pending = counts['pending']
                running = counts['running']
                succeeded = counts['succeeded']
                failed = counts['failed']
                result = f'Pending {pending}\nRunning {running}\nSucceeded {succeeded}\nFailed {failed}'
            else:
                pending, running, succeeded, failed = get_status(cfg.appstate.k8s_ctx, args.run_label, dry_run=dry_run)
                result = f'Pending {pending}\nRunning {running}\nSucceeded {succeeded}\nFailed {failed}'

            if not args.wait or pending + running == 0:
                break
            logging.debug(result)
            time.sleep(20)  # TODO: make this a parameter (granularity)
    except RuntimeError as e:
        returncode = e.args[0]
        print(e.args[1], file=sys.stderr)
    except ValueError as e:
        returncode = 1
        print(e)
    except SafeExecError as err:
        msg = err.message.rstrip().replace('\n', ' | ')
        logging.debug(f'kubectl error: {msg}')
        # if the cluster exists, assume it is initializing
        if gcp.check_cluster(cfg):
            raise UserReportError(CLUSTER_ERROR, f'The cluster "{cfg.cluster.name}" exists, but is not responding. It may be still initializing, please try checking status again in a few minutes.')
        else:
            raise UserReportError(CLUSTER_ERROR, f'The cluster "{cfg.cluster.name}" was not found')
    else:
        if verbose_result:
            print(verbose_result)
        else:
            print(result)
    return returncode



#                           PUBLIC DOMAIN NOTICE
#              National Center for Biotechnology Information
#
# This software is a "United States Government Work" under the
# terms of the United States Copyright Act.  It was written as part of
# the authors' official duties as United States Government employees and
# thus cannot be copyrighted.  This software is freely available
# to the public for use.  The National Library of Medicine and the U.S.
# Government have not placed any restriction on its use or reproduction.
#
# Although all reasonable efforts have been taken to ensure the accuracy
# and reliability of the software and data, the NLM and the U.S.
# Government do not and cannot warrant the performance or results that
# may be obtained by using this software or data.  The NLM and the U.S.
# Government disclaim all warranties, express or implied, including
# warranties of performance, merchantability or fitness for any particular
# purpose.
#
# Please cite NCBI in any work or product based on this material.

"""
elastic_blast/elasticblast.py - Base class for ElasticBlastXXX

Author: Victor Joukov (joukovv@ncbi.nlm.nih.gov)
Created: Tue 03 Aug 2021 06:54:30 PM EDT
"""

from typing import Any, List
from .elb_config import ElasticBlastConfig
from .constants import ElbStatus
from abc import ABCMeta, abstractmethod

class ElasticBlast(metaclass=ABCMeta):
    """ Base class for core ElasticBLAST functionality. """
    def __init__(self, cfg: ElasticBlastConfig, create=False, cleanup_stack: List[Any]=None):
        self.cfg = cfg
        self.cleanup_stack = cleanup_stack if cleanup_stack else []
        self.dry_run = self.cfg.cluster.dry_run

    @abstractmethod
    def cloud_query_split(self, query_files: List[str]) -> None:
        """ Submit the query sequences for splitting to the cloud.
            Parameters:
                query_files - list of files containing query sequence data to split
        """
        self.query_files = query_files

    @abstractmethod
    def wait_for_cloud_query_split(self) -> None:
        """ Wait for cloud query job comletion """
        pass

    @abstractmethod
    def upload_query_length(self, query_length: int) -> None:
        """ Save query length in a metadata file in cloud storage """
        pass

    @abstractmethod
    def check_job_number_limit(self, queries, query_length) -> None:
        """ Check that number of jobs generated does not exceed platform maximum
            If the platform-specific maximum is exceeded throws UserReportError(INPUT_ERROR) """
        pass

    @abstractmethod
    def status(self) -> ElbStatus:
        """ Return the status of an ElasticBLAST search """
        pass

    @abstractmethod
    def delete(self) -> ElbStatus:
        """ Delete all resources allocated for an ElasticBLAST search """
        pass

    @abstractmethod
    def submit(self, query_batches: List[str], one_stage_cloud_query_split: bool) -> None:
        """ Submit query batches to cluster, converts AWS exceptions to UserReportError
            Parameters:
                query_batches               - list of bucket names of queries to submit
                one_stage_cloud_query_split - do the query split in the cloud as a part
                                              of executing a regular job """
        pass



#!/usr/bin/env python3
"""
src/elb/resources/quotas/quota_aws_ec2_cf.py - Module to check AWS resource
availability in EC2 and CloudFormation

Author: Christiam Camacho (camacho@ncbi.nlm.nih.gov)
Created: Mon 14 Sep 2020 10:37:01 AM EDT
"""
import logging
from awslimitchecker.checker import AwsLimitChecker # type: ignore
from elastic_blast.util import UserReportError
from elastic_blast.constants import DEPENDENCY_ERROR


class ResourceCheckAwsEc2CloudFormation:
    """ Class to encapsulate retrieval of EC2 and CloudFormation service quotas, current
    usage and whether this suffices to run ElasticBLAST 
    """

    def __init__(self, boto_cfg = None):
        """ Initialize this object AwsLimitChecker object.
        boto_cfg: boto3 library configuration
        """
        self.checker = AwsLimitChecker() if boto_cfg is None else AwsLimitChecker(region=boto_cfg.region_name)


    def __call__(self):
        """ Retrieve the current usage of the relevant AWS resources and compare it with the service quotas.
        Throws a UserReportError if there aren't enough resources available to run ElasticBLAST
        """
        SERVICES = [ 'EC2', 'CloudFormation' ]
        result = self.checker.check_thresholds(service=SERVICES)
        if not result:
            # No service thresholds were exceeded :)
            return

        fatal_errors = ''
        warnings = ''
        for svc_name in result.keys():
            for usage_metric in result[svc_name].keys():
                if svc_name == 'EC2' and not usage_metric.startswith('Running On-Demand'):
                    continue
                aws_limit = result[svc_name][usage_metric]
                criticals = aws_limit.get_criticals()
                warnings = aws_limit.get_warnings()
                if len(criticals):
                    for c in criticals:
                        fatal_errors += f'{svc_name} metric "{usage_metric}" has reached a critical usage level ({c}) that is too close to the limit ({aws_limit.get_limit()}) to run ElasticBLAST. '
                elif len(warnings):
                    for w in warnings:
                        warnings += f'{svc_name} metric "{usage_metric}" has reached a level of usage ({w}) that is close to the limit ({aws_limit.get_limit()}) and may run into problems. '
        if fatal_errors:
            raise UserReportError(DEPENDENCY_ERROR, fatal_errors)
        if warnings:
            logging.warning(warnings)



#!/usr/bin/env python3
#                           PUBLIC DOMAIN NOTICE
#              National Center for Biotechnology Information
#  
# This software is a "United States Government Work" under the
# terms of the United States Copyright Act.  It was written as part of
# the authors' official duties as United States Government employees and
# thus cannot be copyrighted.  This software is freely available
# to the public for use.  The National Library of Medicine and the U.S.
# Government have not placed any restriction on its use or reproduction.
#   
# Although all reasonable efforts have been taken to ensure the accuracy
# and reliability of the software and data, the NLM and the U.S.
# Government do not and cannot warrant the performance or results that
# may be obtained by using this software or data.  The NLM and the U.S.
# Government disclaim all warranties, express or implied, including
# warranties of performance, merchantability or fitness for any particular
# purpose.
#   
# Please cite NCBI in any work or product based on this material.
"""
create-blastdb-metadata.py - See DESC constant below

Author: Greg Boratyn (boratyng@ncbi.nlm.nih.gov)
Created: Wed 26 May 2020 10:09:23 PM EDT
"""
import os
import argparse
import json
import logging
from pathlib import Path
from elb import VERSION
from elb.util import UserReportError
from elb.util import config_logging, safe_exec

DESC = r"""An application that creates metadata files for BLAST databases"""


def main():
    """ Entry point into this program. """
    METADATA_VERSION = '1.1'

    try:
        parser = create_arg_parser()
        args = parser.parse_args()
        config_logging(args)

        dbname = os.path.basename(args.db)
        metadata = {'version': METADATA_VERSION,
                    'dbname': dbname}

        cmd = 'blastdbcmd -list . -remove_redundant_dbs -list_outfmt'.split() +  ['%f\t%p\t%t\t%l\t%n\t%d\t%U']
        os.chdir('x')
        p = safe_exec(cmd)
#        print(p.stdout.decode())
        
        for line in p.stdout.decode().split('\n'):
            fields = line.rstrip().split('\t')
            if os.path.basename(fields[0]) != dbname:
                continue

            metadata['dbtype'] = fields[1]
            metadata['description'] = fields[2]
            metadata['number-of-letters'] = fields[3]
            metadata['number-of-sequences'] = fields[4]
            metadata['last-updated'] = fields[5]
            metadata['bytes-total'] = fields[6]
            break
        

        if 'description' not in metadata:
            raise UserReportError(returncode=255,
                                  message='Database was not found')

        cmd = f'blastdbcmd -db {dbname} -info'
        p = safe_exec(cmd)
        num_volumes = 0
        
        lines = p.stdout.decode().split('\n')

        for num, line in enumerate(lines):
            if line.startswith('Volumes:'):
                num_volumes = len(lines) - num - 2
                break

        metadata['number-of-volumes'] = num_volumes

        location = '/some/dir'
        files = [f'{location}/{str(i)}' for i in Path('.').glob(dbname + '*')]
        print(files)

        print(json.dumps(metadata, indent=4))

    except UserReportError as err:
        logging.error(err.message)
        return err.returncode

    return 0


def create_arg_parser():
    """ Create the command line options parser object for this script. """
    DFLT_LOGFILE = 'stderr'
    parser = argparse.ArgumentParser(prog=os.path.basename(os.path.splitext(sys.argv[0])[0]), 
                                     description=DESC)
    parser.add_argument("--db", type=str, help="BLAST database to search")
#    parser.add_argument("--db-source", type=str, help="Where NCBI-provided databases are downloaded from", choices=['AWS', 'GCP', 'NCBI'], default='AWS')
    parser.add_argument('--version', action='version',
                        version='%(prog)s ' + VERSION)
    parser.add_argument("--logfile", default=DFLT_LOGFILE, type=str,
                        help="Default: " + DFLT_LOGFILE)
    parser.add_argument("--loglevel", default='INFO',
                        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"])
    return parser


if __name__ == "__main__":
    import sys, traceback
    try:
        sys.exit(main())
    except Exception as e:
        traceback.print_exc(file=sys.stderr)
        sys.exit(1)




#!/usr/bin/env python3
"""
src/elb/resources/quotas/quota_aws_batch.py - Module to check AWS Batch
resource availability

Author: Christiam Camacho (camacho@ncbi.nlm.nih.gov)
Created: Tue 22 Sep 2020 10:37:01 AM EDT
"""
import logging
import boto3 # type: ignore
from pprint import pformat
from botocore.config import Config  # type: ignore
from elastic_blast.util import UserReportError
from elastic_blast.constants import DEPENDENCY_ERROR

OUT_OF_QUOTA_ERR_MSG = 'ElasticBLAST cannot create the necessary AWS resources ({}) to run your search. Please run elastic-blast delete on searches that have already completed.'

class ResourceCheckAwsBatch:
    """ Class to encapsulate retrieval of AWS Batch service quotas, current
    usage and whether this suffices to run ElasticBLAST 
    """
    _service_codes = [ 'batch' ]
    _service_quotas = {
        'Job queue limit': -1,
        'Compute environment limit': -1
    }

    def __init__(self, boto_cfg: Config = None):
        """ Initialize this object with AWS connections and service quotas 
        boto_cfg: boto3 library configuration
        """
        self.client = boto3.client('service-quotas') if boto_cfg is None else boto3.client('service-quotas', config=boto_cfg)
        self.batch = boto3.client('batch') if boto_cfg is None else boto3.client('batch', config=boto_cfg)
        self._initialize_service_quotas()


    def _initialize_service_quotas(self) -> None:
        """ Retrieves the service quotas relevant to ElasticBLAST """
        for svc in self._service_codes:
            done = False
            next_token = ''
            while not done:
                if next_token:
                    response = self.client.list_service_quotas(ServiceCode=svc, NextToken=next_token)
                else:
                    response = self.client.list_service_quotas(ServiceCode=svc)
                for q in response['Quotas']:
                    for quota_name in self._service_quotas.keys():
                        if quota_name == q['QuotaName']:
                            self._service_quotas[quota_name] = q['Value']
                if 'NextToken' in response:
                    next_token = response['NextToken']
                else: 
                    done = True
                    
        logging.debug(f'AWS Batch service quotas: {pformat(self._service_quotas)}')


    def _count_aws_batch_compute_environments(self) -> int:
        """ Count the number of AWS Batch compute environments """
        done = False
        next_token = ''
        count = 0
        while not done:
            if next_token:
                response = self.batch.describe_compute_environments(nextToken=next_token)
            else:
                response = self.batch.describe_compute_environments()
            #logging.debug(f'AWS Batch describe_compute_environments response: {pformat(response)}')
            for ce in response['computeEnvironments']:
                count += 1
            if 'nextToken' in response:
                next_token = response['nextToken']
            else: 
                done = True
        return count


    def _count_aws_batch_job_queues(self) -> int:
        """ Count the number of AWS Batch job queues """
        done = False
        next_token = ''
        count = 0
        while not done:
            if next_token:
                response = self.batch.describe_job_queues(nextToken=next_token)
            else:
                response = self.batch.describe_job_queues()
            #logging.debug(f'AWS Batch describe_job_queues response {pformat(response)}')
            for ce in response['jobQueues']:
                count += 1
            if 'nextToken' in response:
                next_token = response['nextToken']
            else: 
                done = True
        return count


    def __call__(self) -> None:
        """ Retrieve the current usage of the relevant AWS Batch resources and compare it with the service quotas.
        Throws a UserReportError if there aren't enough resources available to run ElasticBLAST
        """
        njq = self._count_aws_batch_job_queues()
        nce = self._count_aws_batch_job_queues()
        logging.debug(f'AWS Batch usage: number of job queues {njq}')
        logging.debug(f'AWS Batch usage: number of compute environments {nce}')
        
        if njq + 1 >= self._service_quotas['Job queue limit']:
            raise UserReportError(DEPENDENCY_ERROR, OUT_OF_QUOTA_ERR_MSG.format('batch job queue'))
        if nce + 1 >= self._service_quotas['Compute environment limit']:
            raise UserReportError(DEPENDENCY_ERROR, OUT_OF_QUOTA_ERR_MSG.format('batch compute environment'))



#                           PUBLIC DOMAIN NOTICE
#              National Center for Biotechnology Information
#  
# This software is a "United States Government Work" under the
# terms of the United States Copyright Act.  It was written as part of
# the authors' official duties as United States Government employees and
# thus cannot be copyrighted.  This software is freely available
# to the public for use.  The National Library of Medicine and the U.S.
# Government have not placed any restriction on its use or reproduction.
#   
# Although all reasonable efforts have been taken to ensure the accuracy
# and reliability of the software and data, the NLM and the U.S.
# Government do not and cannot warrant the performance or results that
# may be obtained by using this software or data.  The NLM and the U.S.
# Government disclaim all warranties, express or implied, including
# warranties of performance, merchantability or fitness for any particular
# purpose.
#   
# Please cite NCBI in any work or product based on this material.

"""
Module status

Reads status of job execution and reports it

Author: Victor Joukov joukovv@ncbi.nlm.nih.gov
"""
import logging
from collections import defaultdict
import re
from . import kubernetes
from .util import safe_exec, SafeExecError, UserReportError
from .constants import CLUSTER_ERROR

#TODO: get rid of unused run_label, EB-407 
def get_status(k8s_ctx: str, run_label=None, dry_run=False):
    """Return numbers of pending, running, succeded, and failed jobs

    run_label: optional argument
        If argument run_label is set it is added to pod selector for kubectl request.
    """
    status = defaultdict(int)  # type: ignore
    selector = 'app=blast'
    # If run_label is set add it to selector
    if run_label:
        # FIXME: EB-132 will need to distinguish this from the the run-label used for cost tracking
        kv = run_label.split(':')
        if len(kv) == 2 and kv[0] and kv[1]:
            selector += f',{kv[0]}={kv[2]}'
        else:
            raise ValueError('Run label not in correct format, must be: <key>:<value>')
    # if we need name of the job in the future add NAME:.metadata.name to custom-columns
    kubectl = f'kubectl --context={k8s_ctx}'

    # get status of jobs (pending/running, succeeded, failed)
    cmd = f'{kubectl} get jobs -o custom-columns=STATUS:.status.conditions[0].type -l {selector}'.split()
    if dry_run:
        logging.debug(cmd)
    else:
        try:
            proc = safe_exec(cmd)
        except SafeExecError as err:
            raise UserReportError(CLUSTER_ERROR, err.message.strip())

        for line in proc.stdout.decode().split('\n'):
            if not line or line.startswith('STATUS'):
                continue
            if line.startswith('Complete'):
                status['Succeeded'] += 1
            elif line.startswith('Failed'):
                status['Failed'] += 1
            else:
                status['Pending'] += 1
            
    # get number of running pods
    cmd = f'{kubectl} get pods -o custom-columns=STATUS:.status.phase -l {selector}'.split()
    if dry_run:
        logging.info(cmd)
    else:
        try:
            proc = safe_exec(cmd)
        except SafeExecError as e:
            raise UserReportError(CLUSTER_ERROR, e.message.strip())
        for line in proc.stdout.decode().split('\n'):
            if line == 'Running':
                status[line] += 1

    # correct number of pending jobs: running jobs were counted twice,
    # as running and pending
    status['Pending'] -= status['Running']

    return status['Pending'], status['Running'], status['Succeeded'], status['Failed']



#!/usr/bin/env python3
"""
create-blastdb-metadata.py - See DESC constant below

Author: Irena Zaretskaya (zaretska@ncbi.nlm.nih.gov)
Created: Fri 4 Sep 2020 02:05:56 PM EST
"""
import os
import time
import argparse
import glob
import configparser
import unittest
import logging
import json
import subprocess
import re
import datetime
from pathlib import Path
import tempfile
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import List, Optional
from tempfile import NamedTemporaryFile


DFLT_LOGFILE = 'create-blastdb-metadata.log'

ERROR_DEPENDENCY = 3
ERROR_DATABASE_INFO = 2
ERROR_DATABASE_NOT_FOUND = 1

DESC = r"""
This program creates BLAST database metadata in JSON format.
"""

"""
Arguments: --db [database]
--logfile [logfile, default:create-blastdb-metadata.log]


Dependencies:
* BLAST+ command line applications


"""

# BLAST database metadata version
BLASTDB_METADATA_VERSION = '1.1'


class DataCollection(Enum):
    """Types of metadata collection from file system"""
    UNCOMPRESSED_FILES = auto()
    CACHED_FILES = auto()


class DbType(Enum):
    """Database molecule type"""
    PROTEIN = 'prot'
    NUCLEOTIDE = 'nucl'

    @classmethod
    def choices(self):
        return [self.PROTEIN.value, self.NUCLEOTIDE.value]


    def __str__(self):
        """Convert value to a string"""
        return self.value


@dataclass
class BlastDbMetadata:
    """Class that holds database metadata information"""
    dbname: str
    version: str = BLASTDB_METADATA_VERSION
    dbtype: str = ''
    description: str = ''
    number_of_letters: int = 0
    number_of_sequences: int = 0
    files: List[str] = field(default_factory=list)
    last_updated: str = ''
    bytes_total: int = 0
    bytes_to_cache: int = 0
    number_of_volumes: int = 0

    def to_json(self, pretty = False):
        """Serialize the object as JSON"""
        output = json.dumps(self, cls=BlastDbMetadataEncoder, sort_keys=False)
        #rename '_' to '-' in keys
        dict = json.loads(output)
        current_keys = list(dict.keys())
        for current_key in current_keys:
            new_key = current_key.replace('_','-')
            dict[new_key] = dict.pop(current_key)
        if pretty:
            output = json.dumps(dict,sort_keys=False, indent=2)
        else:
            output = json.dumps(dict,sort_keys=False)

        return output


class UserReportError(Exception):
    """Exception wich is reported to the user as an error message and exit code"""
    def __init__(self, returncode: int, message: str):
        """Initialize parameters"""
        self.returncode = returncode
        self.message = message

    def __str__(self):
        """Conversion to a string"""
        return self.message


class BlastDbMetadataEncoder(json.JSONEncoder):
    """JSON encoder"""
    def default(self, obj):
        if isinstance(obj, BlastDbMetadata):
            return obj.__dict__
        else:
            return json.JSONEncoder.default(self, obj)


def get_database_info(oneDBJson: BlastDbMetadata, db: Path, dbtype: DbType,
                      file_prefix: Optional[str] = None) -> None:
    """
    Initialize database metadata

    Arguments:
        oneDBJson: object that holds metadata
        db: path plus database name
        dbtype: database molecule type
        file_prefix: if not None, path prefix for database files in metadata
    """
    # get number of volumes
    # this is called first to verify that we are dealing with a proper BLAST
    # database
    cmd = f'blastdbcmd -db {db} -dbtype {dbtype} -info'.split()
    logging.debug(f'Getting number of volumes for {db} CMD: {cmd}')
    try:
        p = subprocess.run(cmd, check=True, stdout = subprocess.PIPE,
                           stderr = subprocess.PIPE)

        lines = p.stdout.decode().split('\n')
        for num, line in enumerate(lines):
            if line.startswith('Volumes:'):
                oneDBJson.number_of_volumes = len(lines) - num - 2
                break

    except subprocess.CalledProcessError as err:
        raise UserReportError(returncode=ERROR_DATABASE_INFO,
                              message=f'{err.stdout.decode().strip()}\n{err.stderr.decode().strip()}')
    except PermissionError:
        raise UserReportError(returncode=ERROR_DEPENDENCY,
                              message='blastdbcmd application could not be executed')


    # get other database information
    # blastdbcmd -list fails for alias files and works with a directory, where
    # additional files may cause it to fail. To avoid problems we create
    # symlinks to database file in a temporary directory and work with them.
    dbfiles = db.parent.glob(f'{db.name}.*[np]??')
    with tempfile.TemporaryDirectory() as tmp:
        logging.debug(f'Creating symlinks to database file in {tmp}')
        # create symlinks, dbfile.resolve() finds an absolute path
        for dbfile in dbfiles:
            (Path(tmp) / dbfile.name).symlink_to(dbfile.resolve())

        cmd = f'blastdbcmd -list {tmp} -dbtype {dbtype} -remove_redundant_dbs -list_outfmt'.split() +  ['%f\t%p\t%t\t%l\t%n\t%d\t%U']
        logging.debug("Getting uncompressed database " + str(db) + " information. CMD: " + ' '.join(cmd))
        try:
            p = subprocess.run(cmd, check=True, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)

            for line in p.stdout.decode().split('\n'):
                fields = line.rstrip().split('\t')
                if os.path.basename(fields[0]) != db.name:
                    continue

                oneDBJson.dbtype = fields[1]
                oneDBJson.description = fields[2]
                oneDBJson.number_of_letters = int(fields[3])
                oneDBJson.number_of_sequences = int(fields[4])
                oneDBJson.last_updated = convert_date_to_iso8601(fields[5])
                oneDBJson.bytes_total = int(fields[6])
                break
        except subprocess.CalledProcessError as err:
            raise UserReportError(returncode=ERROR_DATABASE_INFO,
                                  message=f'{err.stdout.decode().strip()}\n{err.stderr.decode().strip()}')
        except ValueError as err:
            raise UserReportError(returncode=ERROR_DATABASE_INFO,
                                  message=f'Error when parsing database informtation: {err}')
        except PermissionError:
            raise UserReportError(returncode=ERROR_DEPENDENCY,
                                  message='blastdbcmd application could not be executed')

    if not oneDBJson.dbtype:
        # this happens when database is in BLASTDB path, but the path to the
        # database was not provided
        raise UserReportError(returncode=ERROR_DATABASE_INFO,
                              message=f'Database "{db}" was not found. Please, provide path to the database.')

    populate_db_info(oneDBJson, db, DataCollection.UNCOMPRESSED_FILES, file_prefix)
    try:
        populate_db_info(oneDBJson, db, DataCollection.CACHED_FILES, file_prefix)
    except FileNotFoundError as err:
        # if blastdbcmd above worked and database files are missing we are
        # very likely dealing with an alias database
        raise UserReportError(returncode=ERROR_DATABASE_INFO,
                              message=f'{err}. BLAST databases made of alias files that aggregate distinct databases are not supported.')
    logging.debug("Success getting uncompressed database " + str(db) + " information.")

    
def populate_db_info(oneDBJson: BlastDbMetadata, db: Path,
                     data2collect: DataCollection,
                     file_prefix: Optional[str]) -> None:
    """
    Get and populate database information from the filesystem.

    Arguments:
        oneDBJson: object holding database metadata
        db: database path plus name
        data2collect: type of information to gather
        file_prefix: if not None, path prefix for database files in metadata
    """
    dbtype = oneDBJson.dbtype[0].lower()
    if data2collect == DataCollection.UNCOMPRESSED_FILES:
        pattern = '.*' + dbtype + '??'
    elif data2collect == DataCollection.CACHED_FILES:
        pattern = '.*' + dbtype + '[si][qn]'
    else:
        raise NotImplementedError(f'Invalid data collection type {data2collect}')

    fileFound = False       
    
    logging.debug("Searching " + str(db)  + pattern)
    for f in glob.glob(str(db) + pattern):
        if not os.path.islink(f):
            if data2collect == DataCollection.UNCOMPRESSED_FILES:
                filename_parts = os.path.splitext(f)
                if filename_parts[1] == ".md5" or filename_parts[1] == ".gz":
                    continue
                db_file = os.path.basename(f)
                if file_prefix:
                    db_file = os.path.join(file_prefix, db_file)
                oneDBJson.files.append(db_file)
            else:
                oneDBJson.bytes_to_cache += os.stat(f).st_size
            fileFound = True
            
    if not fileFound:
        raise FileNotFoundError(f'Database files: {db}{pattern} not found in {db.parent}')
        
    # include additional CDD files
    # CSeqDB does not know about these files, so their size is not
    # reported by blastdbcmd -list
    for ff in db.parent.glob(db.name + '.*'):
        if data2collect == DataCollection.UNCOMPRESSED_FILES:
            if ff.suffix in ['.aux', '.loo', '.rps', '.freq']:
                db_file = ff.name
                if file_prefix:
                    db_file = os.path.join(file_prefix, ff.name)
                oneDBJson.files.append(db_file)
                oneDBJson.bytes_total += ff.stat().st_size
        else:
            if ff.suffix in ['.loo', '.rps', '.freq']:
                oneDBJson.bytes_to_cache += ff.stat().st_size


def convert_date_to_iso8601(date: str) -> str:
    """Convert a date in human readable format to ISO-8501"""
    months = {'Jan': '01',
              'Feb': '02',
              'Mar': '03',
              'Apr': '04',
              'May': '05',
              'Jun': '06',
              'Jul': '07',
              'Aug': '08',
              'Sep': '09',
              'Oct': '10',
              'Nov': '11',
              'Dec': '12'}

    m = re.findall('([A-Z][a-z]{2})\s(\d{1,2}),\s(\d{4})', date)
    if not m or len(m[0]) < 3:
        raise ValueError(f'The date string "{date}" could not be parsed')

    day = m[0][1]
    month = months[m[0][0]]
    year = m[0][2]

    return f'{year}-{month}-{int(day):02}'


def get_blast_version() -> str:
    """Get BLAST+ version"""
    cmd = 'blastdbcmd -version'.split()
    try:
        p = subprocess.run(cmd, check=True, stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE)

    except subprocess.CalledProcessError as err:
        raise UserReportError(returncode=ERROR_DEPENDENCY,
                              message=f'Unexpected error when checking BLAST+ version')
    except PermissionError:
        raise UserReportError(returncode=ERROR_DEPENDENCY,
                              message='blastdbcmd application could not be executed')

    return p.stdout.decode().split('\n')[0].split()[1][:-1]


def main():
    """ Entry point into this program. """
    parser = create_arg_parser()
    args = parser.parse_args()
        
    config_logging(args)

    db = Path(args.db)
    dbtype = DbType(args.dbtype.lower())
    oneDBJson = BlastDbMetadata(db.name)
    get_database_info(oneDBJson, db, dbtype, args.output_prefix)
    output_file_name = f'{db.name}-{dbtype}-metadata.json'
    if args.out:
        output_file_name = args.out
    with open(output_file_name, 'wt') as output_file:
        print(oneDBJson.to_json(args.pretty), file=output_file)
    logging.debug(f'Metadata printed to {output_file_name}')

    return 0


def create_arg_parser():
    """ Create the command line options parser object for this script. """
    parser = argparse.ArgumentParser(description=DESC)
    parser._action_groups.pop()
    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')

    required.add_argument("--db", metavar='DBNAME', type=str, help="A BLAST database", required=True)
    required.add_argument("--dbtype", type=str,
                        help="Database molecule type", choices=DbType.choices(),
                        required=True)
    optional.add_argument('--out', metavar='FILENAME', help='Output file name. Default: ${db}-${dbtype}-metadata.json')
    optional.add_argument('--output-prefix', metavar='PATH', type=str, help='Path prefix for location of database files in metadata')
    optional.add_argument("--pretty", action='store_true', help="Pretty-print JSON output")
    optional.add_argument("--logfile", default=DFLT_LOGFILE,help="Default: " + DFLT_LOGFILE)
    optional.add_argument("--loglevel", default='INFO',choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"])
    optional.add_argument('--version', action='version', version=f'%(prog)s {get_blast_version()}')
    return parser


def config_logging(args):
    """Configure logging"""
    if args.logfile == 'stderr':
        logging.basicConfig(level=str2ll(args.loglevel),
                            format="%(asctime)s %(message)s")
    else:
        logformat_for_file = "%(asctime)s %(levelname)s: %(message)s"
        logformat_for_stderr = "%(levelname)s: %(message)s"

        logger = logging.getLogger()
        logger.setLevel(str2ll(args.loglevel))

        # to stderr
        handler = logging.StreamHandler()
        handler.setLevel(logging.WARNING)
        handler.setFormatter(logging.Formatter(logformat_for_stderr))
        logger.addHandler(handler)

        # to a file
        handler = logging.FileHandler(args.logfile, mode='a')
        handler.setLevel(str2ll(args.loglevel))
        handler.setFormatter(logging.Formatter(logformat_for_file))
        logger.addHandler(handler)


    logging.logThreads = 0
    logging.logProcesses = 0
    logging._srcfile = None


def str2ll(level):
    """ Converts the log level argument to a numeric value.

    Throws an exception if conversion can't be done.
    Copied from the logging howto documentation
    """
    retval = getattr(logging, level.upper(), None)
    if not isinstance(retval, int):
        raise ValueError('Invalid log level: %s' % level)
    return retval


if __name__ == "__main__":
    import sys
    try:
        sys.exit(main())
    except UserReportError as err:
        logging.error(err.message)
        sys.exit(err.returncode)


import contextlib
import io
import sys
from unittest.mock import patch


class TestCreateDbMetadata(unittest.TestCase):
    "Unit tests for create-blastdb-metadata tool"""

    def test_metadata_version(self):
        """Test that metadata version is set correctly"""
        obj = BlastDbMetadata('some-db')
        json_str = obj.to_json()
        d = json.loads(json_str)
        self.assertEqual(d['version'], BLASTDB_METADATA_VERSION)


    def test_cmd_line_options(self):
        """Test command line parameters parsing"""
        parser = create_arg_parser()

        # minimal parameters
        parser.parse_args(['--db', 'some-db', '--dbtype', 'prot'])

        # --db is required
        with contextlib.redirect_stderr(io.StringIO()) as f:
            with self.assertRaises(SystemExit):
                parser.parse_args(['abc'])
            message = f.getvalue()
            self.assertTrue('required' in message)
            self.assertTrue('--db' in message)
            self.assertTrue('--dbtype' in message)

        # --help and --version should work
        # arparse raises SystemExit withe exit code 0 here
        with contextlib.redirect_stdout(io.StringIO()) as f:
            try:
                parser.parse_args(['--help'])
                parser.parse_args(['--version'])
            except SystemExit:
                pass


    def test_run(self):
        """Test main function with a sample database"""
        # path test database
        db = str(Path(__file__).parent.parent / 'tests/blastdb/testdb')

        with NamedTemporaryFile() as output_file:
            with patch.object(argparse.ArgumentParser, 'parse_args', return_value=argparse.Namespace(db=db, dbtype='prot', out=output_file.name, logfile='stderr', loglevel='ERROR', pretty=None, output_prefix=None)):
                main()
            with open(output_file.name, 'rt') as metadata:
                output = metadata.read()
            result = json.loads(output)
            self.assertEqual(result['version'], BLASTDB_METADATA_VERSION)
            self.assertEqual(result['dbname'], 'testdb')
            self.assertEqual(result['dbtype'], 'Protein')
            self.assertEqual(result['description'], 'Test database: the first four sequences for swissprot')
            self.assertEqual(result['number-of-letters'], 875)
            self.assertEqual(result['number-of-sequences'], 3)
            self.assertEqual(result['last-updated'], '2021-06-17')
            self.assertEqual(result['bytes-total'], 50956)
            self.assertEqual(result['bytes-to-cache'], 1039)
            self.assertEqual(result['number-of-volumes'], 1)
            self.assertEqual(len(result['files']), 9)
            self.assertEqual(len([s for s in result['files'] if s.startswith('testdb.')]), len(result['files']))


    def test_output_prefix(self):
        """Test main function with a sample database and --output-prefix option"""
        # path to test database
        db = str(Path(__file__).parent.parent / 'tests/blastdb/testdb')
        PREFIX = 'gs://some-bucket/some-dir'

        with NamedTemporaryFile() as output_file:
            with patch.object(argparse.ArgumentParser, 'parse_args', return_value=argparse.Namespace(db=db, dbtype='prot', out=output_file.name, logfile='stderr', loglevel='ERROR', pretty=None, output_prefix=PREFIX)):
                main()
            with open(output_file.name, 'rt') as metadata:
                output = metadata.read()
            result = json.loads(output)
            self.assertEqual(len([s for s in result['files'] if s.startswith(PREFIX)]), len(result['files']))


    def test_concert_date_to_iso8601(self):
        """Test converting date to ISO8601 format"""
        self.assertEqual(convert_date_to_iso8601('May 18, 2021  4:21 AM'), '2021-05-18')

        with self.assertRaises(ValueError):
            convert_date_to_iso8601('Some string that is not a date')



#!/usr/bin/env python3
#                           PUBLIC DOMAIN NOTICE
#              National Center for Biotechnology Information
#  
# This software is a "United States Government Work" under the
# terms of the United States Copyright Act.  It was written as part of
# the authors' official duties as United States Government employees and
# thus cannot be copyrighted.  This software is freely available
# to the public for use.  The National Library of Medicine and the U.S.
# Government have not placed any restriction on its use or reproduction.
#   
# Although all reasonable efforts have been taken to ensure the accuracy
# and reliability of the software and data, the NLM and the U.S.
# Government do not and cannot warrant the performance or results that
# may be obtained by using this software or data.  The NLM and the U.S.
# Government disclaim all warranties, express or implied, including
# warranties of performance, merchantability or fitness for any particular
# purpose.
#   
# Please cite NCBI in any work or product based on this material.
"""
blast-tuner.py - See DESC constant below

Author: Christiam Camacho (camacho@ncbi.nlm.nih.gov)
Created: Tue 21 Apr 2020 12:07:42 PM EDT
"""
import os
import argparse
import json
import logging
from elastic_blast import VERSION
from elastic_blast.config import configure
from elastic_blast.util import ElbSupportedPrograms, UserReportError
from elastic_blast.util import get_query_batch_size, config_logging
from elastic_blast.tuner import get_db_data, SeqData, MTMode, MolType
from elastic_blast.tuner import get_mt_mode, get_num_cpus, get_batch_length
from elastic_blast.tuner import aws_get_machine_type, aws_get_mem_limit
from elastic_blast.tuner import gcp_get_machine_type, gcp_get_mem_limit
from elastic_blast.base import DBSource, PositiveInteger, MemoryStr
from elastic_blast.constants import CFG_CLOUD_PROVIDER, CFG_CP_AWS_REGION, CFG_CP_GCP_REGION
from elastic_blast.constants import ELB_DFLT_GCP_REGION, ELB_DFLT_AWS_REGION
from elastic_blast.constants import CFG_BLAST, CFG_CLUSTER
from elastic_blast.constants import CFG_BLAST_DB, CFG_BLAST_PROGRAM, CFG_BLAST_BATCH_LEN
from elastic_blast.constants import CFG_BLAST_OPTIONS, CFG_CLUSTER_NUM_CPUS
from elastic_blast.constants import CFG_CLUSTER_MACHINE_TYPE
from elastic_blast.constants import CFG_BLAST_MEM_LIMIT
from elastic_blast.constants import MolType, INPUT_ERROR, CSP


DESC = r"""This application's purpose is to provide suggestions to run help run
BLAST efficiently. It is meant to be used in conjunction with BLAST+ and
ElasticBLAST."""


def main():
    """ Entry point into this program. """
    try:
        parser = create_arg_parser()
        args = parser.parse_args()
        config_logging(args)
        cfg = configure(args)

        options = '' if args.options is None else args.options
        SECTIONS = [ CFG_CLOUD_PROVIDER, CFG_BLAST, CFG_CLUSTER ]
        conf = { s : {} for s in SECTIONS }
        cloud_provider = CSP[args.csp_target]
        sp = ElbSupportedPrograms()

        db_source = DBSource[cloud_provider.name] if not args.db_source else DBSource[args.db_source]

        if args.db_mem_limit_factor is None:
            db_mem_limit_factor = 0.0 if cloud_provider == CSP.AWS else 1.1
        else:
            db_mem_limit_factor = args.db_mem_limit_factor

        if args.db is not None:
            db_data = get_db_data(args.db, sp.get_db_mol_type(args.program),
                                  db_source)
            conf[CFG_BLAST][CFG_BLAST_DB] = args.db

        if not args.region:
            if cloud_provider == CSP.AWS:
                args.region = ELB_DFLT_AWS_REGION
            else:
                args.region = ELB_DFLT_GCP_REGION

        if cloud_provider == CSP.AWS:
            conf[CFG_CLOUD_PROVIDER][CFG_CP_AWS_REGION] = args.region
        else:
            conf[CFG_CLOUD_PROVIDER][CFG_CP_GCP_REGION] = args.region

        query_data = SeqData(args.total_query_length,
                             sp.get_query_mol_type(args.program))
        mt_mode = get_mt_mode(args.program, args.options, db_data, query_data)
        options += f' {mt_mode}'

        num_cpus = get_num_cpus(program = args.program, mt_mode = mt_mode,
                                query = query_data)
        conf[CFG_BLAST][CFG_BLAST_PROGRAM] = args.program
        conf[CFG_BLAST][CFG_BLAST_BATCH_LEN] = str(get_batch_length(program = args.program,
                                                          mt_mode = mt_mode,
                                                          num_cpus = num_cpus))

        if len(options) > 1:
           conf[CFG_BLAST][CFG_BLAST_OPTIONS] = options

        conf[CFG_CLUSTER][CFG_CLUSTER_NUM_CPUS] = str(num_cpus)

        if args.with_optimal:
            machine_type = 'optimal'
            if cloud_provider != CSP.AWS:
                raise UserReportError(INPUT_ERROR, f'The "optimal" instance type is only allowed for AWS')
        else:
            if cloud_provider == CSP.AWS:
                if args.db == 'nt':
                    machine_type = 'c5ad.16xlarge'
                elif args.db == 'nr':
                    machine_type = 'm5ad.16xlarge'
                else:
                    machine_type = aws_get_machine_type(db = db_data,
                                                        num_cpus = num_cpus,
                                                        region = args.region)
            else:
                machine_type = gcp_get_machine_type(db = db_data,
                                                    num_cpus = num_cpus)

        conf[CFG_CLUSTER][CFG_CLUSTER_MACHINE_TYPE] = machine_type

        if cloud_provider == CSP.AWS:
            mem_limit = aws_get_mem_limit(num_cpus = num_cpus,
                                          machine_type = machine_type,
                                          db = db_data,
                                          db_factor = db_mem_limit_factor)
        else:
            # Currently one job per instance is assumed in GCP
            mem_limit = gcp_get_mem_limit(db = db_data,
                                          db_factor = db_mem_limit_factor)
        conf[CFG_BLAST][CFG_BLAST_MEM_LIMIT] = mem_limit

        for section in SECTIONS:
            print(f'[{section}]', file=args.out)
            for key, value in sorted(conf[section].items(), key=lambda x: x[0]):
                print(f'{key} = {value}', file=args.out)
            print('', file=args.out)

        return 0
    except UserReportError as err:
        logging.error(err.message)
        return err.returncode


def create_arg_parser():
    """ Create the command line options parser object for this script. """
    DFLT_LOGFILE = 'stderr'
    parser = argparse.ArgumentParser(prog=os.path.basename(os.path.splitext(sys.argv[0])[0]), 
                                     description=DESC)
    parser._action_groups.pop()
    required = parser.add_argument_group('required arguments')
    optional = parser.add_argument_group('optional arguments')

    required.add_argument("--db", type=str, help="BLAST database to search", required=True)
    required.add_argument("--program", type=str, help="BLAST program to run",
                        choices=ElbSupportedPrograms().get(), required=True)
    required.add_argument("--total-query-length", type=PositiveInteger, required=True,
                        help='Number of residues or bases in query sequecnes')
    required.add_argument("--csp-target", type=str, help="Which Cloud Service Provider to use, default: AWS", choices=['AWS', 'GCP'], default='AWS')
    required.add_argument('--out', type=argparse.FileType('w'), help='Save configuration in this file', default='-')

    optional.add_argument("--db-source", type=str, help="Where NCBI-provided databases are downloaded from, default: AWS", choices=['AWS', 'GCP', 'NCBI'])
    optional.add_argument("--region", type=str, help=f'Cloud Service Provider region. Defaults: {ELB_DFLT_AWS_REGION} for AWS; {ELB_DFLT_GCP_REGION} for GCP')
    optional.add_argument("--options", type=str, help='BLAST options', default='')
    optional.add_argument("--db-mem-limit-factor", type=float,
                          help='This number times database bytes-to-cache will produce memory limit for a BLAST search. (default: 0.0: for AWS, 1.1 for GCP)')
    optional.add_argument("--with-optimal", action='store_true',
                         help='Use AWS optimal instance type')
    optional.add_argument('--version', action='version',
                        version='%(prog)s ' + VERSION)
    optional.add_argument("--logfile", default=DFLT_LOGFILE, type=str,
                        help="Default: " + DFLT_LOGFILE)
    optional.add_argument("--loglevel", default='INFO',
                        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"])
    return parser


if __name__ == "__main__":
    import sys, traceback
    try:
        sys.exit(main())
    except Exception as e:
        traceback.print_exc(file=sys.stderr)
        sys.exit(1)




#                           PUBLIC DOMAIN NOTICE
#              National Center for Biotechnology Information
#  
# This software is a "United States Government Work" under the
# terms of the United States Copyright Act.  It was written as part of
# the authors' official duties as United States Government employees and
# thus cannot be copyrighted.  This software is freely available
# to the public for use.  The National Library of Medicine and the U.S.
# Government have not placed any restriction on its use or reproduction.
#   
# Although all reasonable efforts have been taken to ensure the accuracy
# and reliability of the software and data, the NLM and the U.S.
# Government do not and cannot warrant the performance or results that
# may be obtained by using this software or data.  The NLM and the U.S.
# Government disclaim all warranties, express or implied, including
# warranties of performance, merchantability or fitness for any particular
# purpose.
#   
# Please cite NCBI in any work or product based on this material.
"""
elb/tuner.py - Functions for checking and estimating ElasticBLAST compute and
memory requirements

Created: Fri 14 May 2021 05:10:18 PM EDT
"""

import json, logging, os
from enum import Enum
from dataclasses import dataclass
from bisect import bisect_left
import math
from .filehelper import open_for_read
from .constants import BLASTDB_ERROR, INPUT_ERROR, ELB_BLASTDB_MEMORY_MARGIN
from .constants import UNKNOWN_ERROR, MolType
from .constants import ELB_S3_PREFIX, ELB_GCS_PREFIX
from .constants import SYSTEM_MEMORY_RESERVE
from .base import DBSource, PositiveInteger, MemoryStr
from .util import UserReportError, get_query_batch_size
from .aws_traits import get_instance_type_offerings, get_suitable_instance_types
from .aws_traits import get_machine_properties as aws_get_machine_properties
from .gcp_traits import GCP_MACHINES

@dataclass
class SeqData:
    """Basic sequence information"""
    length: int
    moltype: MolType


@dataclass
class DbData(SeqData):
    """Database information"""
    bytes_to_cache_gb: float


class MTMode(Enum):
    """Values for BLAST MT mode option

    Str converter generates commad line option."""
    ZERO = 0
    ONE = 1

    def __str__(self):
        """Convert enum value to BLAST commad line options string"""
        return '-mt_mode 1' if self.value == 1 else ''


# MT mode 1 query length per thread
# From https://preview.ncbi.nlm.nih.gov/books/prevqa/NBK322759/ , in particular
# As noted above, the threading by query option works well with relatively small databases if you have a lot of queries to process.
# For BLASTP, your input FASTA should contain at least 10,000 residues per thread, or 320,000 residues for 32 threads. 
# For BLASTX, the FASTA should contain 30,000 bases per thread. For both BLASTP and BLASTX, we find threading by query works best 
# if the database has fewer than 2 billion residues (e.g, swissprot). For BLASTN, your FASTA file should contain at least 2.5 
# million bases per thread and the database should contain fewer than 14 billion residues. Threading by query also works well if
# you limit the search, such as by -taxid or -gilist, regardless of the size of the database.
RESIDUES_PER_THREAD = 10000
BASES_PER_THREAD = 2500000
BASES_PER_THREAD_BLASTX = 30000

# Maximum database length for MT mode 1
MAX_MT_ONE_DB_LENGTH_PROT = 2000000000
MAX_MT_ONE_DB_LENGTH_NUCL = 14000000000

# Number of threads for MT mode 0
NUM_THREADS_MT_ZERO = 16


def get_db_data(db: str, dbtype: MolType, source: DBSource) -> DbData:
    """
    Read database metadata.

    Arguments:
        db: Database name or URI
        dbtype: Molecule type for BLASTDB
        source: Source for NCBI provided database, ignored for a user database
    """
    DB_BUCKET_AWS = os.path.join(ELB_S3_PREFIX, 'ncbi-blast-databases')
    DB_BUCKET_GCP = os.path.join(ELB_GCS_PREFIX, 'blast-db')
    DB_BUCKET_NCBI = 'ftp://ftp.ncbi.nlm.nih.gov/blast/db'

    # for user databases
    metadata_file = f"{db}-{dbtype}-metadata.json"

    # if an NCBI-provided database
    if not db.startswith(ELB_S3_PREFIX) and not db.startswith(ELB_GCS_PREFIX):
        if source == DBSource.AWS or source == DBSource.GCP:
            bucket = DB_BUCKET_AWS if source == DBSource.AWS else DB_BUCKET_GCP
            with open_for_read(f'{bucket}/latest-dir') as f:
                metadata_file = os.path.join(f'{bucket}/{f.readline().rstrip()}', metadata_file)
        else:
            metadata_file = os.path.join(f'{DB_BUCKET_NCBI}', metadata_file)
    logging.debug(f'BLASTDB metadata file: {metadata_file}')
    try:
        with open_for_read(metadata_file) as f:
            db_metadata = json.load(f)
    except:
        msg = f'Metadata file for database: "{db}" does not exist in {source}'
        if db.startswith(ELB_S3_PREFIX) or db.startswith(ELB_GCS_PREFIX):
            msg += ' or you lack credentials to access this file'
        msg += '.'
        raise UserReportError(returncode=BLASTDB_ERROR, message=msg)

    return DbData(length = int(db_metadata['number-of-letters']),
                  moltype = MolType[db_metadata['dbtype'].upper()],
                  bytes_to_cache_gb = int(db_metadata['bytes-to-cache']) / (1024 ** 3))


def get_mt_mode(program: str, options: str, db: DbData, query: SeqData) -> MTMode:
    """
    Compute BLAST search MT mode

    Arguments:
        program: BLAST program
        options: BLAST options (empty string for defaults)
        db: Database information
        query: Queries information
    """
    if program == 'blastx' and query.length <= BASES_PER_THREAD_BLASTX:
        return MTMode.ZERO
    if (query.moltype == MolType.PROTEIN and query.length <= RESIDUES_PER_THREAD) or \
       (query.moltype == MolType.NUCLEOTIDE and query.length <= BASES_PER_THREAD):
        return MTMode.ZERO

    if program.lower() == 'rpsblast' or program.lower() == 'rpstblastn':
        return MTMode.ONE

    if '-taxids' in options or '-taxidlist' in options or \
           '-gilist' in options or '-seqidlist' in options or \
           '-negative_taxids' in options or '-negative_taxidlist' in options:
        return MTMode.ONE

    if (db.moltype == MolType.PROTEIN and db.length <= MAX_MT_ONE_DB_LENGTH_PROT) or \
       (db.moltype == MolType.NUCLEOTIDE and db.length <= MAX_MT_ONE_DB_LENGTH_NUCL):
        return MTMode.ONE

    return MTMode.ZERO


def get_num_cpus(program: str, mt_mode: MTMode, query: SeqData) -> int:
    """Get number of CPUs to use to optimally run BLAST

    Arguments:
        program: BLAST program
        mt_mode: BLAST MT mode
        query: Queries information"""
    if mt_mode == MTMode.ZERO:
        return NUM_THREADS_MT_ZERO
    else:
        if program.lower() == 'blastx':
            characters_per_thread = BASES_PER_THREAD_BLASTX
        else:
            characters_per_thread = RESIDUES_PER_THREAD if query.moltype == MolType.PROTEIN else BASES_PER_THREAD
        return min([query.length // characters_per_thread + 1, 16])


def get_batch_length(program: str, mt_mode: MTMode, num_cpus: int) -> int:
    """
    Get batch length for BLAST batch search

    Arguments:
        program: BLAST program
        mt_mode: BLAST MT mode
        num_cpus: Number of threads/CPUs to use
    """
    batch_len = get_query_batch_size(program)
    if batch_len is None:
        raise UserReportError(returncode=INPUT_ERROR,
                              message=f"Invalid BLAST program '{program}'")

    if program.lower == 'blastx':
        batch_len = BASES_PER_THREAD_BLASTX

    if mt_mode == MTMode.ONE:
        batch_len *= num_cpus

    return batch_len


def aws_get_mem_limit(num_cpus: PositiveInteger, machine_type: str,
                      db: DbData, db_factor: float) -> MemoryStr:
    """Get memory limit for searching a single query batch for AWS

    Arguments:
        num_cpus: Number of CPUs per search job
        db: Database information
        machine_type: Machine type (can be "optimali")
        db_factor: If larger than 0, memory limit will be computed as databse
                   bytes-to-cache times db_factor

    Returns:
        A search job memory limit as MemoryStr"""
    if db_factor > 0.0:

        return MemoryStr(f'{round(db.bytes_to_cache_gb * db_factor, 1)}G')
    elif machine_type == 'optimal':
        result = 60 if db.bytes_to_cache_gb >= 60 else db.bytes_to_cache_gb + 2
        return MemoryStr(f'{result}G')
    else:
        props = aws_get_machine_properties(machine_type)
        jobs_per_instance = math.floor(props.ncpus / num_cpus)
        mem_limit = round((props.memory - SYSTEM_MEMORY_RESERVE) / jobs_per_instance, 1)
        return MemoryStr(f'{mem_limit}G')


def gcp_get_mem_limit(db: DbData, db_factor: float) -> MemoryStr:
    """Get memory limit for searching a single query batch for GCP. One job
    per instance is assumed.

    Arguments:
        db: Database information
        db_factor: If larger than 0, memory limit will be computed as database
                   bytes-to-cache times db_factor

    Returns:
        A search job memory limit as MemoryStr"""
    return MemoryStr(f'{round(db.bytes_to_cache_gb * db_factor, 1)}G')


def aws_get_machine_type(db: DbData, num_cpus: PositiveInteger, region: str) -> str:
    """Select an AWS machine type that can accommodate the database and has
    enough VCPUs. The machine type is selected from a set of machines supported
    by AWS Batch and offered in a specific region.

    Arguments:
        db: Database information
        num_cpus: Required number of CPUs
        region: Cloud provided region

    Returns:
        Instance type with at least as much memory as db.bytes_to_cache_gb and
        required number of CPUs"""
    M5_FAMILY = ['m5ad.large', 'm5ad.xlarge'] + [f'm5ad.{i}xlarge' for i in [2, 4, 8, 12, 16, 24]]
    C5_FAMILY = ['c5ad.large', 'c5ad.xlarge'] + [f'c5ad.{i}xlarge' for i in [2, 4, 8, 12, 16, 24]]
    R5_FAMILY = ['r5ad.large', 'r5ad.xlarge'] + [f'r5ad.{i}xlarge' for i in [2, 4, 8, 12, 16, 24]]
    SUPPORTED_INSTANCE_TYPES = M5_FAMILY + C5_FAMILY + R5_FAMILY

    # get a list of instance types offered in the region
    offerings = get_instance_type_offerings(region)

    supported_offerings = [tp for tp in offerings if tp in SUPPORTED_INSTANCE_TYPES]

    # get properties of suitable instances
    suitable_props = get_suitable_instance_types(min_memory=MemoryStr(f'{db.bytes_to_cache_gb * ELB_BLASTDB_MEMORY_MARGIN}G'),
                                                 min_cpus=num_cpus,
                                                 instance_types=supported_offerings)

    return sorted(suitable_props, key=lambda x: x['MemoryInfo']['SizeInMiB'])[0]['InstanceType']


def gcp_get_machine_type(db: DbData, num_cpus: PositiveInteger) -> str:
    """Select a GCP machine type that can accommodate the database and has
    enough VCPUs. The instance type is selected from E2 and N1 instance
    families. One job per instance is assumed.

    Arguments:
        db: Database metadata
        num_cpus: Required number of CPUs

    Returns:
        Instance type with at least as much memory as mem_limit and
        required number of CPUs"""
    # machine type families ranked by memory to CPU ratio (order matters)
    FAMILIES_N1 = ['n1-highcpu',
                   'n1-standard',
                   'n1-highmem']

    FAMILIES_E2 = ['e2-standard', 'e2-highmem']

    # numbers of CPUs in GCP instance types
    CPUS = [1, 2, 4, 8, 16, 32, 64, 96]

    idx = bisect_left(CPUS, num_cpus)
    if idx >= len(CPUS):
        raise UserReportError(returncode=UNKNOWN_ERROR,
                              message=f'GCP machine type with {num_cpus} CPUs could not be found')

    memory = db.bytes_to_cache_gb + SYSTEM_MEMORY_RESERVE

    # E2 machine types are usually the cheapest, but have at most 128GB memory
    machine_type_families = FAMILIES_E2 if memory < 120 and num_cpus <= 32 else FAMILIES_N1

    machine_cpus = CPUS[idx]
    mem_cpu_ratio = memory / machine_cpus
    family = None
    while idx < len(CPUS) and not family:
        machine_cpus = CPUS[idx]
        mem_cpu_ratio = memory / machine_cpus
        for fam in machine_type_families:
            if mem_cpu_ratio <= GCP_MACHINES[fam]:
                family = fam
                break
        if not family:
            idx += 1

    if not family:
        raise UserReportError(returncode=UNKNOWN_ERROR,
                              message=f'GCP machine type with memory {memory}GB and {num_cpus} CPUs could not be found')

    return f'{family}-{machine_cpus}'



#!/usr/bin/env python3
"""
bin/elastic-blast.py - See DESC constant below

Author: Christiam Camacho (camacho@ncbi.nlm.nih.gov)
Created: Wed 22 Apr 2020 06:31:30 AM EDT
"""
import os
import sys
import signal
import argparse
import logging
from typing import List
from elb import VERSION
from elb.commands.submit import submit as elb_submit
from elb.commands.status import create_arg_parser as create_status_arg_parser
from elb.commands.delete import delete as elb_delete
from elb.commands.run_summary import create_arg_parser as create_run_summary_arg_parser
from elb.util import check_positive_int, config_logging, UserReportError, SafeExecError
from elb.util import ElbSupportedPrograms, clean_up
from elb import constants
from elb.gcp import check_prerequisites
from elb.config import configure
from elb.elb_config import ElasticBlastConfig


DFLT_LOGFILE = 'elastic-blast.log'
DESC = r"""This application facilitates running BLAST on large amounts of query sequence data
on the cloud"""

# error message for missing Elastic-BLAST task on the command line
NO_TASK_MSG =\
"""Elastic-BLAST task was not specified. Please, use submit, status, delete, or run-summary.
usage: elastic-blast [-h] [--version] {submit,status,delete,run-summary} --cfg <config file> [options]"""

def main():
    """Local main entry point which sets up arguments, undo stack,
    and processes exceptions """
    try:
        signal.signal(signal.SIGINT, signal.default_int_handler)
        clean_up_stack = []
        # Check parameters for Unicode letters and reject if codes higher than 255 occur
        reject_cli_args_with_unicode(sys.argv[1:])
        parser = create_arg_parser()
        args = parser.parse_args()
        if not args.subcommand:
            # report missing command line task
            raise UserReportError(returncode=constants.INPUT_ERROR,
                                  message=NO_TASK_MSG)
        config_logging(args)
        cfg = configure(args)
        logging.info(f"ElasticBLAST {args.subcommand} {VERSION}")
        logging.debug({s: dict(cfg[s]) for s in cfg.sections()})
        check_prerequisites(ElasticBlastConfig(cfg))
        #TODO: use cfg only when args.wait, args.sync, and args.run_label are replicated in cfg
        return args.func(args, cfg, clean_up_stack)
    except (SafeExecError, UserReportError) as e:
        logging.error(e.message)
        # SafeExecError return code is the exit code from command line
        # application ran via subprocess
        if isinstance(e, SafeExecError):
            return constants.DEPENDENCY_ERROR
        return e.returncode
    except KeyboardInterrupt:
        return constants.INTERRUPT_ERROR
    #TODO: process filehelper.TarReadError here
    finally:
        messages = clean_up(clean_up_stack)
        if messages:
            for msg in messages:
                logging.error(msg)
            sys.exit(constants.UNKNOWN_ERROR)


def reject_string_with_unicode(content: str) -> None:
    for c in content:
        if ord(c) > 255:
            raise UserReportError(returncode=constants.INPUT_ERROR,
                                  message=f"Command line has Unicode letters in argument '{content}', can't be processed")


def reject_cli_args_with_unicode(args: List[str]) -> None:
        for arg in args:
            reject_string_with_unicode(arg)


def file_must_exist(path: str) -> str:
    """Check if given  path exists and is a file, helper function for
    argparse.ArgumentParser. If used for a command line argument, the
    application will exit with an error if file is not found or path is not a
    file."""
    if os.path.isfile(path):
        return path
    raise argparse.ArgumentTypeError(f'File {path} was not found')


def positive_int(arg: str) -> int:
    """Positive integer type for argparse.ArgumentParser. Raises
    argparse.ArgumentTypeError if the supplied string is not a positive integer."""
    retval = None
    try:
        retval = check_positive_int(arg)
    except Exception as err:
        raise argparse.ArgumentTypeError(str(err))
    return retval


def create_arg_parser():
    """ Create the command line options parser object for this script. """
    prog = os.path.splitext(os.path.basename(sys.argv[0]))[0]
    parser = ElbArgumentParser(prog=prog, description=DESC,
        epilog="To get help about specific command run %(prog)s command --help")
    parser.add_argument('--version', action='version', version='%(prog)s ' + VERSION)

    common_opts_parser = ElbArgumentParser(add_help=False)
    common_opts_parser.add_argument("--cfg", metavar="FILE",
                                    type=file_must_exist,
                                    help="ElasticBLAST configuration file")
    common_opts_parser.add_argument(f"--{constants.CFG_BLAST_RESULTS}", type=str,
                        help="Bucket URI where to save the output from ElasticBLAST")
    common_opts_parser.add_argument("--logfile", default=DFLT_LOGFILE, type=str,
                                    help="Default: " + DFLT_LOGFILE)
    common_opts_parser.add_argument("--loglevel", default='INFO',
                                    choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"])
    common_opts_parser.add_argument("--dry-run", action='store_true', 
                                    help="Do not perform any actions")

    sp = parser.add_subparsers(dest='subcommand')#title='Subcommands')
    create_submit_arg_parser(sp, common_opts_parser)
    create_status_arg_parser(sp, common_opts_parser)
    create_delete_arg_parser(sp, common_opts_parser)
    create_run_summary_arg_parser(sp, common_opts_parser)
    return parser


def create_submit_arg_parser(subparser, common_opts_parser):
    """ Create the command line options subparser for the submit command. """
    parser = subparser.add_parser('submit', help='Submit an ElasticBLAST search',
                                  parents=[common_opts_parser])
    parser.add_argument("--program", type=str, help="BLAST program to run",
                        choices=ElbSupportedPrograms().get())
    parser.add_argument("--query", type=str,
                        help="Query sequence data, can be provided as a local path or GCS bucket URI to a single file/tarball")
    parser.add_argument("--db", type=str, help="BLAST database to search")
    parser.add_argument("--num-nodes", type=positive_int,
                        help="Number of worker nodes to use")
    #parser.add_argument("--sync", action='store_true', 
    #                    help="Run in synchronous mode")
    # FIXME: EB-132
    parser.add_argument("--run-label", type=str,
                        help="Run-label to tag this ElasticBLAST search, format: key:value")
    parser.add_argument('blast_opts', nargs=argparse.REMAINDER,
                        metavar='BLAST_OPTS',
                        help="Options to pass to BLAST program")
    parser.set_defaults(func=elb_submit)
    return parser


def create_delete_arg_parser(subparser, common_opts_parser):
    """ Create the command line options subparser for the status command. """
    parser = subparser.add_parser('delete',
                                  parents=[common_opts_parser],
                                  help='Delete resources associated with an ElasticBLAST search')
    # FIXME: EB-132
    parser.add_argument("--run-label", type=str,
                        help="Run-label for the ElasticBLAST search to delete, format: key:value")
    parser.set_defaults(func=elb_delete)


class ElbArgumentParser(argparse.ArgumentParser):
    """Custom argument parser to override application exit code"""
    def exit(self, status=0, message=None):
        """Custom exit function that overrides ArgumentParser application
        exit code"""
        if status:
            super().exit(constants.INPUT_ERROR, message)
        else:
            super().exit()

    def error(self, message):
        """Custom error message that does not print usage on errors"""
        self.exit(constants.INPUT_ERROR, f'{self.prog}: error: {message}\n')


if __name__ == "__main__":
    sys.exit(main())
    import traceback
    try:
        sys.exit(main())
    except Exception as e:
        traceback.print_exc(file=sys.stderr)
        sys.exit(constants.UNKNOWN_ERROR)



#!/usr/bin/env python3
#                           PUBLIC DOMAIN NOTICE
#              National Center for Biotechnology Information
#
# This software is a "United States Government Work" under the
# terms of the United States Copyright Act.  It was written as part of
# the authors' official duties as United States Government employees and
# thus cannot be copyrighted.  This software is freely available
# to the public for use.  The National Library of Medicine and the U.S.
# Government have not placed any restriction on its use or reproduction.
#   
# Although all reasonable efforts have been taken to ensure the accuracy
# and reliability of the software and data, the NLM and the U.S.
# Government do not and cannot warrant the performance or results that
# may be obtained by using this software or data.  The NLM and the U.S.
# Government disclaim all warranties, express or implied, including
# warranties of performance, merchantability or fitness for any particular
# purpose.
#   
# Please cite NCBI in any work or product based on this material.

# Script to submit ElasticBLAST jobs by an AWS Batch job
#
# Author: Christiam Camacho camacho@ncbi.nlm.nih.gov


import argparse
import logging
import os
from botocore.exceptions import ClientError # type: ignore
from tempfile import NamedTemporaryFile
from pathlib import Path
from pprint import pformat
from elastic_blast.constants import ElbCommand, ELB_DFLT_LOGLEVEL, ElbStatus
from elastic_blast.constants import CSP, ELB_S3_PREFIX
from elastic_blast.constants import ELB_METADATA_DIR, ELB_STATUS_SUCCESS, ELB_STATUS_FAILURE
from elastic_blast.elasticblast import ElasticBlast
from elastic_blast.aws import ElasticBlastAws
from elastic_blast.elb_config import ElasticBlastConfig
from elastic_blast.util import config_logging

from elastic_blast.filehelper import thaw_config, upload_file_to_gcs, check_for_read
from elastic_blast.object_storage_utils import copy_file_to_s3


DESC = 'ElasticBLAST Janitor module to clean up after itself'

def copy_to_results_bucket_if_not_present(filename: str, bucket: str):
    """ Wrapper function to copy a file to cloud object storage """
    try:
        check_for_read(bucket)
    except FileNotFoundError:
        if bucket.startswith(ELB_S3_PREFIX):
            copy_file_to_s3(bucket, Path(filename))
        else:
            upload_file_to_gcs(filename, bucket)


def janitor(elb: ElasticBlast) -> None:
    """ ElasticBLAST Janitor function: cleans up ElasticBLAST resources """
    st = elb.status()
    results = elb.cfg.cluster.results
    cluster_name = elb.cfg.cluster.name
    if st == ElbStatus.SUCCESS:
        with NamedTemporaryFile() as f:
            copy_to_results_bucket_if_not_present(f.name, os.path.join(results, ELB_METADATA_DIR, ELB_STATUS_SUCCESS))
        logging.debug(f'ElasticBLAST search with results on {results} is DONE, deleting it (cluster name {cluster_name})')
        elb.delete()
    elif st == ElbStatus.FAILURE:
        with NamedTemporaryFile() as f:
            copy_to_results_bucket_if_not_present(f.name, os.path.join(results, ELB_METADATA_DIR, ELB_STATUS_FAILURE))
        logging.debug(f'ElasticBLAST search with results on {results} has FAILED, deleting it (cluster name {cluster_name})')
        elb.delete()
    elif st == ElbStatus.RUNNING:
        logging.debug(f'ElasticBLAST search with results on {results} is still running (cluster name {cluster_name})')
    elif st == ElbStatus.UNKNOWN:
        logging.warning(f'Unknown or expired ElasticBLAST search with results on {results} (cluster name {cluster_name})')



def main():
    """Main function, provided for testing """
    parser = create_arg_parser()
    args = parser.parse_args()
    config_logging(args)
    try:
        cf = thaw_config(args.results)
        #logging.debug(f'Recovered configuration from {args.results}: {pformat(cf)}')
        if 'aws' in cf:
            logging.debug(f'AWS region: {cf["aws"]["region"]}')
        else:
            logging.debug(f'GCP project: {cf["gcp"]["project"]}')
            logging.debug(f'GCP region: {cf["gcp"]["region"]}')
            logging.debug(f'GCP zone: {cf["gcp"]["zone"]}')
        logging.debug(f'BLAST program: {cf["blast"]["program"]}')
        logging.debug(f'BLAST database: {cf["blast"]["db"]}')
        logging.debug(f'ELB_CLUSTER_NAME: {cf["cluster"]["name"]}')
        logging.debug(f'ELB_RESULTS: {cf["cluster"]["results"]}')
        # FIXME: only AWS is supported right now
        cfg = ElasticBlastConfig(aws_region = cf['aws']['region'],
                                 program = cf['blast']['program'],
                                 db = cf['blast']['db'],
                                 queries = 'fake',
                                 results = args.results,
                                 cluster_name = cf['cluster']['name'],
                                 task = ElbCommand.STATUS)
        cfg.validate(ElbCommand.STATUS)
        #logging.debug(f'{pformat(cfg.asdict())}')

        if 'ELB_CLUSTER_NAME' in os.environ:
            cfg.cluster.name = os.environ['ELB_CLUSTER_NAME']

        eb = ElasticBlastAws(cfg)
        eb.dry_run = args.dry_run
        janitor(eb)
    except ClientError as ex:
        if ex.response['Error']['Code'] == 'NoSuchKey':
            logging.fatal(f'Results bucket {args.results} does not contain an ElasticBLAST configuration')
        return 1
    return 0


def create_arg_parser():
    """ Create the command line options parser object for this script. """
    parser = argparse.ArgumentParser(description=DESC)
    parser.add_argument('--results', metavar='STR', type=str, help='Results bucket', required=True)
    parser.add_argument("--dry-run", action='store_true', help="Do not perform any actions")
    parser.add_argument("--logfile", default='stderr', type=str, help=f"Default: stderr")
    parser.add_argument("--loglevel", default=ELB_DFLT_LOGLEVEL,
                        help=f"Default: {ELB_DFLT_LOGLEVEL}",
                        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"])
    return parser

if __name__ == '__main__':
    import sys, traceback
    try:
        sys.exit(main())
    except Exception as e:
        print(e, file=sys.stderr)
        traceback.print_exc(file=sys.stderr)
        sys.exit(1)


# vim: set syntax=python ts=4 et :



#                           PUBLIC DOMAIN NOTICE
#              National Center for Biotechnology Information
#  
# This software is a "United States Government Work" under the
# terms of the United States Copyright Act.  It was written as part of
# the authors' official duties as United States Government employees and
# thus cannot be copyrighted.  This software is freely available
# to the public for use.  The National Library of Medicine and the U.S.
# Government have not placed any restriction on its use or reproduction.
#   
# Although all reasonable efforts have been taken to ensure the accuracy
# and reliability of the software and data, the NLM and the U.S.
# Government do not and cannot warrant the performance or results that
# may be obtained by using this software or data.  The NLM and the U.S.
# Government disclaim all warranties, express or implied, including
# warranties of performance, merchantability or fitness for any particular
# purpose.
#   
# Please cite NCBI in any work or product based on this material.

"""
elb/comands/delete.py - delete cluster

Author: Victor Joukov joukovv@ncbi.nlm.nih.gov
"""

from elastic_blast.aws import ElasticBlastAws
from elastic_blast.gcp import delete_cluster_with_cleanup, enable_gcp_api
from elastic_blast.constants import CSP, ElbCommand
from elastic_blast.elb_config import ElasticBlastConfig

# TODO: use cfg only when args.wait, args.sync, and args.run_label are replicated in cfg
def delete(args, cfg, clean_up_stack):
    """ Entry point to delete resources associated with an ElasticBLAST search """
    cfg.validate(ElbCommand.DELETE)
    if cfg.cloud_provider.cloud == CSP.AWS:
        elastic_blast = ElasticBlastAws(cfg)
        elastic_blast.delete()
    else:
        enable_gcp_api(cfg)
        delete_cluster_with_cleanup(cfg)
    return 0



#                           PUBLIC DOMAIN NOTICE
#              National Center for Biotechnology Information
#
# This software is a "United States Government Work" under the
# terms of the United States Copyright Act.  It was written as part of
# the authors' official duties as United States Government employees and
# thus cannot be copyrighted.  This software is freely available
# to the public for use.  The National Library of Medicine and the U.S.
# Government have not placed any restriction on its use or reproduction.
#
# Although all reasonable efforts have been taken to ensure the accuracy
# and reliability of the software and data, the NLM and the U.S.
# Government do not and cannot warrant the performance or results that
# may be obtained by using this software or data.  The NLM and the U.S.
# Government disclaim all warranties, express or implied, including
# warranties of performance, merchantability or fitness for any particular
# purpose.
#
# Please cite NCBI in any work or product based on this material.

"""
Help functions to access GCP resources and manipulate parameters and environment

Authors: Greg Boratyn boratyng@ncbi.nlm.nih.gov
         Yuriy Merezhuk merezhuk@ncbi.nlm.nih.gov
"""

import os
from pathlib import Path
from tempfile import TemporaryDirectory
import time
import logging
import json
from timeit import default_timer as timer
from typing import Any, Optional, List
import uuid
from tenacity import retry, stop_after_attempt, wait_exponential

from .filehelper import open_for_write_immediate
from .jobs import read_job_template, write_job_files
from .util import ElbSupportedPrograms, get_blastdb_info, get_usage_reporting, safe_exec, UserReportError, SafeExecError
from .util import validate_gcp_disk_name

from . import kubernetes
from .constants import CLUSTER_ERROR, GCP_APIS, ELB_METADATA_DIR, ELB_LOG_DIR
from .constants import ELB_STATE_DISK_ID_FILE, CSP, DEPENDENCY_ERROR
from .constants import ELB_QUERY_BATCH_DIR, ELB_DFLT_MIN_NUM_NODES
from .constants import K8S_JOB_BLAST, K8S_JOB_GET_BLASTDB, K8S_JOB_IMPORT_QUERY_BATCHES
from .constants import K8S_JOB_LOAD_BLASTDB_INTO_RAM, K8S_JOB_RESULTS_EXPORT, K8S_UNINITIALIZED_CONTEXT
from .constants import ELB_DOCKER_IMAGE_GCP, ELB_QUERY_LENGTH, INPUT_ERROR
from .constants import ElbStatus
from .elb_config import ElasticBlastConfig
from .elasticblast import ElasticBlast

class ElasticBlastGcp(ElasticBlast):
    """ Implementation of core ElasticBLAST functionality in GCP. """
    def __init__(self, cfg: ElasticBlastConfig, create=False, cleanup_stack: List[Any]=None):
        super().__init__(cfg, create, cleanup_stack)
        self.query_files: List[str] = []
        self.cluster_initialized = False

    def cloud_query_split(self, query_files: List[str]) -> None:
        """ Submit the query sequences for splitting to the cloud.
            Parameters:
                query_files     - list files containing query sequence data to split
        """
        self.query_files = query_files

    def wait_for_cloud_query_split(self) -> None:
        """ Initialize cluster with cloud split job """
        if self.dry_run:
            return
        self._initialize_cluster()
        self.cluster_initialized = True

    def upload_query_length(self, query_length: int) -> None:
        """ Save query length in a metadata file in GS """
        if query_length <= 0: return
        fname = os.path.join(self.cfg.cluster.results, ELB_METADATA_DIR, ELB_QUERY_LENGTH)
        with open_for_write_immediate(fname) as f:
            f.write(str(query_length))
        # Note: if cloud split is used this file is uploaded
        # by the run script in the 1st stage

    def check_job_number_limit(self, queries, query_length) -> None:
        """ Check that resulting number of jobs does not exceed Kubernetes limit """
        k8s_job_limit = kubernetes.get_maximum_number_of_allowed_k8s_jobs(self.dry_run)
        if len(queries) > k8s_job_limit:
            batch_len = self.cfg.blast.batch_len
            suggested_batch_len = int(query_length / k8s_job_limit) + 1
            msg = f'The batch size specified ({batch_len}) led to creating {len(queries)} kubernetes jobs, which exceeds the limit on number of jobs ({k8s_job_limit}).' \
                  f' Please increase the batch-len parameter to at least {suggested_batch_len}.'
            raise UserReportError(INPUT_ERROR, msg)

    def submit(self, query_batches: List[str], one_stage_cloud_query_split: bool) -> None:
        """ Submit query batches to cluster, converts AWS exceptions to UserReportError
            Parameters:
                query_batches               - list of bucket names of queries to submit
                one_stage_cloud_query_split - do the query split in the cloud as a part
                                              of executing a regular job """
        # Can't use one stage cloud split for GCP, should never happen
        assert(not one_stage_cloud_query_split)
        if not self.cluster_initialized:
            self.query_files = []
            self._initialize_cluster()
            self.cluster_initialized = True
        self._generate_and_submit_jobs(query_batches)
        # Sync mode disabled per EB-700
        #if args.sync:
        #    while True:
        #        try:
        #            pending, running, succeeded, failed = get_status(args.run_label, dry_run=dry_run)
        #        except RuntimeError as e:
        #            returncode = e.args[0]
        #            logging.error(f'Error while getting job status: {e.args[1]}, returncode: {returncode}')
        #            # TODO: maybe analyze situation in more details here. It happens when kubectl can't be found
        #            # or cluster connection can't be established. If the latter, maybe try to get GKE credentials again
        #        except ValueError as e:
        #            returncode = 1
        #            logging.error(f'Error while getting job status: {e}')
        #            # This error happens when run-label is malformed, it will not repair, so exit here
        #            break
        #        else:
        #            if pending + running:
        #                logging.debug(f'Pending {pending}, Running {running}, Succeeded {succeeded}, Failed {failed}')
        #            else:
        #                logging.info(f'Done: {succeeded} jobs succeeded, {failed} jobs failed')
        #                break
        #        time.sleep(20)  # TODO: make this a parameter (granularity)
        #    logging.info('Deleting cluster')
        #else:
        self.cleanup_stack.clear()
        self.cleanup_stack.append(lambda: kubernetes.collect_k8s_logs(self.cfg))


    def _initialize_cluster(self):
        """ Creates a k8s cluster, connects to it and initializes the persistent disk """
        cfg, query_files, clean_up_stack = self.cfg, self.query_files, self.cleanup_stack
        logging.info('Starting cluster')
        clean_up_stack.append(lambda: logging.debug('Before creating cluster'))
        clean_up_stack.append(lambda: delete_cluster_with_cleanup(cfg))
        clean_up_stack.append(lambda: kubernetes.collect_k8s_logs(cfg))
        start_cluster(cfg)
        clean_up_stack.append(lambda: logging.debug('After creating cluster'))

        cfg.appstate.k8s_ctx = get_gke_credentials(cfg)

        logging.info('Initializing storage')
        clean_up_stack.append(lambda: logging.debug('Before initializing storage'))
        kubernetes.initialize_storage(cfg, query_files)
        clean_up_stack.append(lambda: logging.debug('After initializing storage'))


    def _generate_and_submit_jobs(self, queries: List[str]):
        cfg, clean_up_stack = self.cfg, self.cleanup_stack
        usage_reporting = get_usage_reporting()

        db, _, db_label = get_blastdb_info(cfg.blast.db)

        # Job generation
        blast_program = cfg.blast.program

        # prepare substitution for current template
        # TODO consider template using cfg variables directly as, e.g. ${blast.program}
        subs = {
            'ELB_BLAST_PROGRAM': blast_program,
            'ELB_DB': db,
            'ELB_DB_LABEL': db_label,
            'ELB_MEM_REQUEST': str(cfg.blast.mem_request),
            'ELB_MEM_LIMIT': str(cfg.blast.mem_limit),
            'ELB_BLAST_OPTIONS': cfg.blast.options,
            # FIXME: EB-210
            'ELB_BLAST_TIMEOUT': str(cfg.timeouts.blast_k8s * 60),
            'BUCKET': cfg.cluster.results,
            'ELB_NUM_CPUS': str(cfg.cluster.num_cpus),
            'ELB_DB_MOL_TYPE': str(ElbSupportedPrograms().get_db_mol_type(blast_program)),
            'ELB_DOCKER_IMAGE': ELB_DOCKER_IMAGE_GCP,
            'ELB_TIMEFMT': '%s%N',  # timestamp in nanoseconds
            'BLAST_ELB_JOB_ID': uuid.uuid4().hex,
            'BLAST_USAGE_REPORT': str(usage_reporting).lower(),
            'K8S_JOB_GET_BLASTDB' : K8S_JOB_GET_BLASTDB,
            'K8S_JOB_LOAD_BLASTDB_INTO_RAM' : K8S_JOB_LOAD_BLASTDB_INTO_RAM,
            'K8S_JOB_IMPORT_QUERY_BATCHES' : K8S_JOB_IMPORT_QUERY_BATCHES,
            'K8S_JOB_BLAST' : K8S_JOB_BLAST,
            'K8S_JOB_RESULTS_EXPORT' : K8S_JOB_RESULTS_EXPORT
        }

        job_template_text = read_job_template(cfg=cfg)
        with TemporaryDirectory() as job_path:
            job_files = write_job_files(job_path, 'batch_', job_template_text, queries, **subs)
            logging.debug(f'Generated {len(job_files)} job files')
            if len(job_files) > 0:
                logging.debug(f'Job #1 file: {job_files[0]}')
                logging.debug('Command to run in the pod:')
                with open(job_files[0]) as f:
                    for line in f:
                        if line.find('-query') >= 0:
                            logging.debug(line.strip())
                            break

            logging.info('Submitting jobs to cluster')
            clean_up_stack.append(lambda: logging.debug('Before submission computational jobs'))
            # Should never happen, cfg.appstate.k8s_ctx should always be initialized properly
            # by the time of this call 
            assert(cfg.appstate.k8s_ctx)
            job_names = kubernetes.submit_jobs(cfg.appstate.k8s_ctx, Path(job_path), dry_run=self.dry_run)
            clean_up_stack.append(lambda: logging.debug('After submission computational jobs'))
            if job_names:
                logging.debug(f'Job #1 name: {job_names[0]}')

        # Allow this to permit invocation of run-summary before deletion of all metadata
        if 'ELB_DISABLE_AUTO_SHUTDOWN' not in os.environ:
            kubernetes.submit_janitor_cronjob(cfg)

    def status(self) -> ElbStatus:
        pass

    def delete(self):
        pass

def enable_gcp_api(cfg: ElasticBlastConfig):
    """ Enable GCP APIs if they are not already enabled 
    parameters:
        cfg: configuration object
    raises:
        SafeExecError if there is an error checking or trying to enable APIs
    """
    dry_run = cfg.cluster.dry_run
    for api in GCP_APIS:
        cmd = 'gcloud services list --enabled --format=value(config.name) '
        cmd += f'--filter=config.name={api}.googleapis.com '
        cmd += f'--project {cfg.gcp.project}'
        if dry_run:
            logging.info(cmd)
        else:
            p = safe_exec(cmd)
            if not p.stdout:
                cmd = f'gcloud services enable {api}.googleapis.com '
                cmd += f'--project {cfg.gcp.project}'
                p = safe_exec(cmd)


def get_gcp_project() -> Optional[str]:
    """Return current GCP project or None if the property is unset.

    Raises:
        util.SafeExecError on problems with command line gcloud
        RuntimeError if gcloud run is successful, but the result is empty"""
    cmd: str = 'gcloud config get-value project'
    p = safe_exec(cmd)
    result: Optional[str]

    # the result should not be empty, for unset properties gcloud returns the
    # string: '(unset)' to stderr
    if not p.stdout and not p.stderr:
        raise RuntimeError('Current GCP project could not be established')

    result = p.stdout.decode().split('\n')[0]

    # return None if project is unset
    if result == '(unset)':
        result = None
    return result


def set_gcp_project(project: str) -> None:
    """Set current GCP project in gcloud environment, raises
    util.SafeExecError on problems with running command line gcloud"""
    cmd = f'gcloud config set project {project}'
    safe_exec(cmd)


def get_disks(cfg: ElasticBlastConfig, dry_run: bool = False) -> List[str]:
    """Return a list of disk names in the current GCP project.
    Raises:
        util.SafeExecError on problems with command line gcloud,
        RuntimeError when gcloud results cannot be parsed"""
    cmd = f'gcloud compute disks list --format json --project {cfg.gcp.project}'
    if dry_run:
        logging.info(cmd)
        return list()

    p = safe_exec(cmd)
    try:
        disks = json.loads(p.stdout.decode())
    except Exception as err:
        raise RuntimeError('Error when parsing listing of GCP disks' + str(err))
    if disks is None:
        raise RuntimeError('Improperly read gcloud disk listing')
    return [i['name'] for i in disks]


def delete_disk(name: str, cfg: ElasticBlastConfig) -> None:
    """Delete a persistent disk.

    Arguments:
        name: Disk name
        cfg: Application config

    Raises:
        util.SafeExecError on problems with command line tools
        ValueError if disk name is empty"""
    if not name:
        raise ValueError('No disk name provided')
    if not cfg:
        raise ValueError('No application config provided')
    cmd = f'gcloud compute disks delete -q {name} --project {cfg.gcp.project}  --zone {cfg.gcp.zone}'
    safe_exec(cmd)


@retry(reraise=True, stop=stop_after_attempt(3), wait=wait_exponential(multiplier=1, min=2, max=10))
def _get_pd_id(cfg: ElasticBlastConfig) -> List[str]:
    """ Try to get the GCP persistent disk ID from elastic-blast records"""
    retval = list()
    if cfg.appstate.disk_id:
        retval = [cfg.appstate.disk_id]
        logging.debug(f'GCP disk ID {retval[0]}')
        # no need to get disk id from GS if we already have it
        return retval

    disk_id_on_gcs = os.path.join(cfg.cluster.results, ELB_METADATA_DIR, ELB_STATE_DISK_ID_FILE)
    cmd = f'gsutil -q stat {disk_id_on_gcs}'
    try:
        safe_exec(cmd)
    except Exception as e:
        logging.debug(f'{disk_id_on_gcs} not found')
        return retval

    cmd = f'gsutil -q cat {disk_id_on_gcs}'
    try:
        p = safe_exec(cmd)
        gcp_disk_id = p.stdout.decode().strip('\n')
        err = p.stderr.decode()
        if gcp_disk_id:
            logging.debug(f"Retrieved GCP disk ID {gcp_disk_id} from {disk_id_on_gcs}")
            try:
                validate_gcp_disk_name(gcp_disk_id)
            except ValueError:
                logging.error(f'GCP disk ID "{gcp_disk_id}" retrieved from {disk_id_on_gcs} is invalid.')
                gcp_disk_id = ''
            else:
                retval.append(gcp_disk_id)
        else:
            raise RuntimeError('Persistent disk id stored in GS is empty')
    except Exception as e:
        logging.error(f'Unable to read {disk_id_on_gcs}: {e}')
        raise

    logging.debug(f'Fetched disk IDs {retval}')
    return retval


def delete_cluster_with_cleanup(cfg: ElasticBlastConfig) -> None:
    """Delete GKE cluster along with persistent disk

    Arguments:
        cfg: Config parameters"""

    dry_run = cfg.cluster.dry_run
    try_kubernetes = True
    pds = []
    try:
        pds = _get_pd_id(cfg)
    except Exception as e:
        logging.error(f'Unable to read disk id from GS: {e}')
    else:
        logging.debug(f'PD id {" ".join(pds)}')

    # detrmine the course of action based on cluster status
    while True:
        status = check_cluster(cfg)
        if not status:
            msg = f'Cluster {cfg.cluster.name} was not found'
            if cfg.cluster.dry_run:
                logging.error(msg)
                return
            else:
                raise UserReportError(returncode=CLUSTER_ERROR, message=msg)
        logging.debug(f'Cluster status "{status}"')

        if status == 'RUNNING' or status == 'RUNNING_WITH_ERROR':
            break
        # if error, there is something wrong with the cluster, kubernetes will
        # likely not work
        if status == 'ERROR':
            try_kubernetes = False
            break
        # if cluster is provisioning or undergoing software updates, wait
        # until it is active,
        if status == 'PROVISIONING' or status == 'RECONCILING':
            time.sleep(10)
            continue
        # if cluster is already being deleted, nothing to do, exit with an error
        if status == 'STOPPING':
            raise UserReportError(returncode=CLUSTER_ERROR,
                                  message=f"cluster '{cfg.cluster.name}' is already being deleted")

        # for unrecognized cluster status exit the loop and the code below
        # will delete the cluster
        logging.warning(f'Unrecognized cluster status {status}')
        break

    if try_kubernetes:
        try:
            cfg.appstate.k8s_ctx = get_gke_credentials(cfg)
            kubernetes.check_server(cfg.appstate.k8s_ctx, dry_run)
        except Exception as e:
            logging.warning(f'Connection to Kubernetes cluster failed.\tDetails: {e}')
            # Can't do anything kubernetes without cluster credentials
            try_kubernetes = False

    if try_kubernetes:
        k8s_ctx = cfg.appstate.k8s_ctx
        # This should never happen when calling the elastic-blast script, as
        # the k8s context is set as part of calling gcloud container clusters get credentials
        # This check is to pacify the mypy type checker and to alert those
        # using the API directly of missing pre-conditions
        assert(k8s_ctx)

        try:
            # get cluster's persistent disk in case they leak
            pds = kubernetes.get_persistent_disks(k8s_ctx, dry_run)
        except Exception as e:
            logging.warning(f'kubernetes.get_persistent_disks failed.\tDetails: {e}')

        try:
            # delete all k8s jobs, persistent volumes and volume claims
            # this should delete persistent disks
            deleted = kubernetes.delete_all(k8s_ctx, dry_run)
            logging.debug(f'Deleted k8s objects {" ".join(deleted)}')
            disks = get_disks(cfg, dry_run)
            for i in pds:
                if i in disks:
                    logging.debug(f'PD {i} still present after deleting k8s jobs and PVCs')
                else:
                    logging.debug(f'PD {i} was deleted by deleting k8s PVC')
        except Exception as e:
            # nothing to do the above fails, the code below will take care of
            # persistent disk leak
            logging.warning(f'kubernetes.delete_all failed.\tDetails: {e}')

    if pds:
        try:
            # delete persistent disks if they are still in GCP, this may be faster
            # than deleting a non-existent disk
            disks = get_disks(cfg, dry_run)
            for i in pds:
                if i in disks:
                    logging.debug(f'PD {i} still present after cluster deletion, deleting again')
                    delete_disk(i, cfg)
        except Exception as e:
            logging.error(getattr(e, 'message', repr(e)))

            # if the above failed, try deleting each disk unconditionally to
            # minimize resource leak
            for i in pds:
                try:
                    delete_disk(i, cfg)
                except Exception as e:
                    logging.error(getattr(e, 'message', repr(e)))
        finally:
            disks = get_disks(cfg, dry_run)
            for i in pds:
                if i in disks:
                    msg = f'ElasticBLAST was not able to delete persistent disk "{i}". Leaving  it may cause additional charges from the cloud provider. You can verify that the disk still exists using this command:\ngcloud compute disks list --project {cfg.gcp.project} | grep {i}\nand delete it with:\ngcloud compute disks delete {i} --project {cfg.gcp.project}'
                    logging.error(msg)
                    # Remove the exception for now, as we want to delete the cluster always!
                    #raise UserReportError(returncode=CLUSTER_ERROR, msg)

    remove_split_query(cfg)
    delete_cluster(cfg)
    remove_ancillary_data(cfg, ELB_LOG_DIR)
    remove_ancillary_data(cfg, ELB_METADATA_DIR)


def get_gke_clusters(cfg: ElasticBlastConfig) -> List[str]:
    """Return a list of GKE cluster names.

    Arguments:
        cfg: configuration object

    Raises:
        util.SafeExecError on problems with command line gcloud
        RuntimeError on problems parsing gcloud JSON output"""
    cmd = f'gcloud container clusters list --format json --project {cfg.gcp.project}'
    p = safe_exec(cmd)
    try:
        clusters = json.loads(p.stdout.decode())
    except Exception as err:
        raise RuntimeError(f'Error when parsing JSON listing of GKE clusters: {str(err)}')
    return [i['name'] for i in clusters]


def get_gke_credentials(cfg: ElasticBlastConfig) -> str:
    """Connect to a GKE cluster.

    Arguments:
        cfg: configuration object

    Returns:
        The kubernetes current context

    Raises:
        util.SafeExecError on problems with command line gcloud"""
    cmd: List[str] = 'gcloud container clusters get-credentials'.split()
    cmd.append(cfg.cluster.name)
    cmd.append('--project')
    cmd.append(f'{cfg.gcp.project}')
    cmd.append('--zone')
    cmd.append(f'{cfg.gcp.zone}')
    if cfg.cluster.dry_run:
        logging.info(cmd)
    else:
        safe_exec(cmd)

    cmd = 'kubectl config current-context'.split()
    retval = K8S_UNINITIALIZED_CONTEXT
    if cfg.cluster.dry_run:
        logging.info(cmd)
    else:
        p = safe_exec(cmd)
        retval = p.stdout.decode()
    return retval


def check_cluster(cfg: ElasticBlastConfig):
    """ Check if cluster specified by configuration is running.
    Returns cluster status - RUNNING, PROVISIONING, STOPPING, or ERROR -
    if there is such cluster, empty string otherwise.
    All possible exceptions will be passed to upper level.
    """
    cluster_name = cfg.cluster.name
    # FIXME: Consider using
    # gcloud container clusters describe CLUSTER_NAME --format value(status) --project=PROJECT --zone=ZONE
    # The difference is that for non-existing name it will throw an exception, but for long cluster list
    # can be faster (depends on gcloud implementation)
    cmd = f'gcloud container clusters list --format=value(status) --filter name={cluster_name} --project {cfg.gcp.project}'
    retval = ''
    if cfg.cluster.dry_run:
        logging.info(cmd)
    else:
        out = safe_exec(cmd)
        retval = out.stdout.decode('utf-8').strip()
    return retval


def start_cluster(cfg: ElasticBlastConfig):
    """ Starts cluster as specified by configuration.
    All possible exceptions will be passed to upper level.

    Per https://cloud.google.com/kubernetes-engine/docs/how-to/creating-a-regional-cluster#create-regional-single-zone-nodepool
    this function creates a (standard GKE) regional cluster with a single-zone node pool
    """

    cluster_name = ''
    machine_type = ''
    num_nodes = 1

    # .. get values from config and raise exception if missing
    if cfg.cluster.name is not None:
        cluster_name = cfg.cluster.name
    else:
        raise ValueError('Configuration error: missing cluster name in [cluster] sections')
    if cfg.cluster.machine_type is not None:
        machine_type = cfg.cluster.machine_type
    else:
        raise ValueError('Configuration error: missing machine-type in [cluster] sections')
    if cfg.cluster.num_nodes is not None:
        num_nodes = cfg.cluster.num_nodes
    else:
        raise ValueError('Configuration error: missing num-nodes in [cluster] sections')

    # ask for cheaper nodes
    use_preemptible = cfg.cluster.use_preemptible
    use_local_ssd = cfg.cluster.use_local_ssd
    dry_run = cfg.cluster.dry_run

    actual_params = ["gcloud", "container", "clusters", "create", cluster_name]
    actual_params.append('--project')
    actual_params.append(f'{cfg.gcp.project}')
    actual_params.append('--zone')
    actual_params.append(f'{cfg.gcp.zone}')

    actual_params.append('--machine-type')
    actual_params.append(machine_type)

    actual_params.append('--num-nodes')
    actual_params.append(str(ELB_DFLT_MIN_NUM_NODES))
    # Autoscaling configuration
    actual_params.append('--enable-autoscaling')
    actual_params.append('--min-nodes')
    actual_params.append(str(ELB_DFLT_MIN_NUM_NODES))
    actual_params.append('--max-nodes')
    actual_params.append(str(num_nodes))

    if use_preemptible:
        actual_params.append('--preemptible')

    # https://cloud.google.com/stackdriver/pricing
    if cfg.cluster.enable_stackdriver:
        actual_params.append('--enable-stackdriver-kubernetes')

    # ATT: hardcoded parameters
    # specifies GCP API in use
    actual_params.append('--scopes')
    scopes = 'compute-rw,storage-rw,cloud-platform,logging-write,monitoring-write'
    actual_params.append(scopes)

    # FIXME: labels, in future will be provided by config or run-time
    labels = cfg.cluster.labels
    actual_params.append('--labels')
    actual_params.append(labels)

    if use_local_ssd:
        actual_params.append('--local-ssd-count')
        actual_params.append('1')

    if cfg.gcp.network is not None:
        actual_params.append(f'--network={cfg.gcp.network}')
    if cfg.gcp.subnet is not None:
        actual_params.append(f'--subnetwork={cfg.gcp.subnet}')

    start = timer()
    if dry_run:
        logging.info(' '.join(actual_params))
    else:
        safe_exec(actual_params)
    end = timer()
    logging.debug(f'RUNTIME cluster-create {end-start} seconds')

    # FIXME: this code needs to be moved outside this function, after
    # get_gke_credentials is called so the proper --context kubectl 
    # argument can be used
    if use_local_ssd:
        # Label nodes in the cluster for affinity
        cmd = "kubectl get nodes -o jsonpath={.items[*]['metadata.name']}"
        if dry_run:
            logging.info(cmd)
            res = ' '.join([f'gke-node-{i}' for i in range(num_nodes)])
        else:
            proc = safe_exec(cmd)
            res = proc.stdout.decode()
        for i, name in enumerate(res.split()):
            cmd = f'kubectl label nodes {name} ordinal={i}'
            if dry_run:
                logging.info(cmd)
            else:
                safe_exec(cmd)
    return cluster_name


def delete_cluster(cfg: ElasticBlastConfig):
    cluster_name = cfg.cluster.name
    actual_params = ["gcloud", "container", "clusters", "delete", cluster_name]
    actual_params.append('--project')
    actual_params.append(f'{cfg.gcp.project}')
    actual_params.append('--zone')
    actual_params.append(f'{cfg.gcp.zone}')
    actual_params.append('--quiet')
    start = timer()
    if cfg.cluster.dry_run:
        logging.info(actual_params)
    else:
        safe_exec(actual_params)
    end = timer()
    logging.debug(f'RUNTIME cluster-delete {end-start} seconds')
    return cluster_name


def check_prerequisites(cfg: ElasticBlastConfig) -> None:
    """ Check that necessary tools, gcloud, gsutil, and kubectl
    are available if necessary.
    If execution of one of these tools is unsuccessful
    it will throw UserReportError exception."""
    if cfg.cloud_provider.cloud == CSP.GCP:
        try:
            safe_exec('gcloud --version')
        except SafeExecError as e:
            message = f"Required pre-requisite 'gcloud' doesn't work, check installation of GCP SDK.\nDetails: {e.message}"
            raise UserReportError(DEPENDENCY_ERROR, message)
        try:
            # client=true prevents kubectl from addressing server which can be down at the moment
            safe_exec('kubectl version --client=true')
        except SafeExecError as e:
            message = f"Required pre-requisite 'kubectl' doesn't work, check Kubernetes installation.\nDetails: {e.message}"
            raise UserReportError(DEPENDENCY_ERROR, message)
    # FIXME: query files supplied in query-list files should be checked too
    # when EB-780 is done
    # For status checking and cluster deletion blast config is not required.
    queries = cfg.blast.queries_arg if cfg.blast else None
    if (queries and len([i for i in queries.split() if i.startswith('gs://')])) or \
           cfg.cluster.results.startswith('gs://'):
        # Check we have gsutil available
        try:
            safe_exec('gsutil --version')
        except SafeExecError as e:
            message = f"Required pre-requisite 'gsutil' doesn't work, check installation of GCP SDK.\nDetails: {e.message}\nNote: this is because your query is located on GS, you may try another location"
            raise UserReportError(DEPENDENCY_ERROR, message)


def remove_split_query(cfg: ElasticBlastConfig) -> None:
    """ Remove split query from user's results bucket """
    remove_ancillary_data(cfg, ELB_QUERY_BATCH_DIR)


def remove_ancillary_data(cfg: ElasticBlastConfig, bucket_prefix: str) -> None:
    """ Removes ancillary data from the end user's result bucket
    cfg: Configuration object
    bucket_prefix: path that follows the users' bucket name (looks like a file system directory)
    """
    dry_run = cfg.cluster.dry_run
    if cfg.cloud_provider.cloud == CSP.GCP:
        out_path = os.path.join(cfg.cluster.results, bucket_prefix, '*')
        cmd = f'gsutil -mq rm {out_path}'
        if dry_run:
            logging.info(cmd)
        else:
            # This command is a part of clean-up process, there is no benefit in reporting
            # its failure except logging it
            try:
                safe_exec(cmd)
            except SafeExecError as e:
                message = e.message.strip().translate(str.maketrans('\n', '|'))
                logging.warning(message)
    elif cfg.cloud_provider.cloud == CSP.AWS:
        # TODO: implement clean up of S3 bucket from split queries here. EB-508
        pass



#                           PUBLIC DOMAIN NOTICE
#              National Center for Biotechnology Information
#  
# This software is a "United States Government Work" under the
# terms of the United States Copyright Act.  It was written as part of
# the authors' official duties as United States Government employees and
# thus cannot be copyrighted.  This software is freely available
# to the public for use.  The National Library of Medicine and the U.S.
# Government have not placed any restriction on its use or reproduction.
#   
# Although all reasonable efforts have been taken to ensure the accuracy
# and reliability of the software and data, the NLM and the U.S.
# Government do not and cannot warrant the performance or results that
# may be obtained by using this software or data.  The NLM and the U.S.
# Government disclaim all warranties, express or implied, including
# warranties of performance, merchantability or fitness for any particular
# purpose.
#   
# Please cite NCBI in any work or product based on this material.

"""
Module elastic_blast.subst - substitute variables of form ${VAR_NAME} from either parameters or environment

Author: Victor Joukov joukovv@ncbi.nlm.nih.gov
"""
import re

re_sub = re.compile(r'\$(?:\{([A-Za-z_][A-Za-z0-9_]*)\}|([A-Za-z_][A-Za-z0-9_]*))')
def substitute_params(job_template: str, map_obj) -> str:
    """ Substitute variables ${QUERY}, ${QUERY_PATH}, ... ${RESULTS} with
    actual values to form a valid YAML job file.
    Also substitute variables from OS environment

    Params:

    job_template: text to substitute variables in
    map_obj: object with get method to use for substitutions

    Returns: text with substitutions
    """
    def _subs_var(mo):
        v = ''
        if mo.group(1):
            v = mo.group(1)
        else:
            v = mo.group(2)
        return map_obj.get(v, mo.group(0))
    return re_sub.sub(_subs_var, job_template)

